/**
 *  '$RCSfile: AddMissedSubTreeSAXHandler.java,v $'
 *    Purpose: A Class that handles the SAX XML events as they
 *             are generated from XML documents
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones, Jivka Bojilova
 *    Release: 1.5.0
 *
 *   '$Author: tao $'
 *     '$Date: 2003/04/02 17:58:39 $'
 * '$Revision: 1.1 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.sql.*;
import java.io.File;
import java.io.FileWriter;
import java.io.StringReader;
import java.util.Stack;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.EmptyStackException;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ext.DeclHandler;
import org.xml.sax.ext.LexicalHandler;
import org.xml.sax.helpers.DefaultHandler;

/**
 * A database aware Class implementing callback bethods for the SAX parser to
 * call when processing the XML stream and generating events
 */
public class AddMissedSubTreeSAXHandler extends DefaultHandler
{
   String    docId;
   Hashtable unReadableSubTreeHash;
   Vector    unReadableSubTreeList = new Vector();
   
   // Constant
   private static final String DESCRIBES = "describes";
   private static final String ADDITIONALMETADATA = "additionalMetadata";
   private static final String ORDER = "order";
   private static final String ID ="id";
   private static final String REFERENCES = "references";
   public  static final String INLINE = "inline";
      
  
    /** Construct an instance of the handler class
    * In this constructor, user can specify the version need to upadate
    *
    * @param myDocid  the docid for these subtree
    * @param myUnReadableTree  the tree couldn't be read
    */
   public AddMissedSubTreeSAXHandler(Hashtable myUnReadableTree) 
                                                           throws SAXException
   {
     unReadableSubTreeHash = myUnReadableTree;
     getUnReadableSubTreeList();
   }
   
   /* Get the unreadable tree list (every tree has the noderecorder stack)*/
   private void getUnReadableSubTreeList() throws SAXException
   {
     Enumeration en = unReadableSubTreeHash.elements();
     
     try
     {
       // go though the unreadable subtree list
       while (en.hasMoreElements())
       {
         // Get  a subtree without node record list
         SubTree treeWithoutStack = (SubTree)en.nextElement();
         String subTreeId   = treeWithoutStack.getSubTreeId();
         long   startNodeId = treeWithoutStack.getStartNodeId();
         long   endNodeId   = treeWithoutStack.getEndNodeId();
         // Get a subtree has the nodelist
         SubTree tree       = new SubTree(docId, subTreeId, 
                                          startNodeId, endNodeId);
         // add this tree to the vector
         unReadableSubTreeList.add(tree);
       }//while
     }//try
     catch (Exception e)
     {
       throw new SAXException(e.getMessage());
     }
   }
   
   /** SAX Handler that is called at the start of each XML element */
   public void startElement(String uri, String localName,
                            String qName, Attributes atts)
               throws SAXException 
  {
    
  }
  
   
  /** SAX Handler that is called at the end of each XML element */
   public void endElement(String uri, String localName,
                          String qName) throws SAXException 
  {
     MetaCatUtil.debugMessage("End ELEMENT " + qName, 50);

   
   }
   
}
