/**
 *  '$RCSfile: DBQuery.java,v $'
 *    Purpose: A Class that searches a relational DB for elements and
 *             attributes that have free text matches a query string,
 *             or structured query matches to a path specified node in the
 *             XML hierarchy.  It returns a result set consisting of the
 *             document ID for each document that satisfies the query
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: 1.5.0
 *
 *   '$Author: sgarg $'
 *     '$Date: 2005/04/13 18:38:16 $'
 * '$Revision: 1.111 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.io.StringWriter;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.servlet.ServletOutputStream;
import javax.servlet.http.HttpServletResponse;


import edu.ucsb.nceas.morpho.datapackage.Triple;
import edu.ucsb.nceas.morpho.datapackage.TripleCollection;

/**
 * A Class that searches a relational DB for elements and attributes that have
 * free text matches a query string, or structured query matches to a path
 * specified node in the XML hierarchy. It returns a result set consisting of
 * the document ID for each document that satisfies the query
 */
public class DBQuery
{

    static final int ALL = 1;

    static final int WRITE = 2;

    static final int READ = 4;

    //private Connection conn = null;
    private String parserName = null;

    private MetaCatUtil util = new MetaCatUtil();

    /**
     * the main routine used to test the DBQuery utility.
     * <p>
     * Usage: java DBQuery <xmlfile>
     *
     * @param xmlfile the filename of the xml file containing the query
     */
    static public void main(String[] args)
    {

        if (args.length < 1) {
            System.err.println("Wrong number of arguments!!!");
            System.err.println("USAGE: java DBQuery [-t] [-index] <xmlfile>");
            return;
        } else {
            try {

                int i = 0;
                boolean showRuntime = false;
                boolean useXMLIndex = false;
                if (args[i].equals("-t")) {
                    showRuntime = true;
                    i++;
                }
                if (args[i].equals("-index")) {
                    useXMLIndex = true;
                    i++;
                }
                String xmlfile = args[i];

                // Time the request if asked for
                double startTime = System.currentTimeMillis();

                // Open a connection to the database
                MetaCatUtil util = new MetaCatUtil();
                //Connection dbconn = util.openDBConnection();

                double connTime = System.currentTimeMillis();

                // Execute the query
                DBQuery queryobj = new DBQuery(MetaCatUtil
                        .getOption("saxparser"));
                FileReader xml = new FileReader(new File(xmlfile));
                Hashtable nodelist = null;
                //nodelist = queryobj.findDocuments(xml, null, null, useXMLIndex);

                // Print the reulting document listing
                StringBuffer result = new StringBuffer();
                String document = null;
                String docid = null;
                result.append("<?xml version=\"1.0\"?>\n");
                result.append("<resultset>\n");

                if (!showRuntime) {
                    Enumeration doclist = nodelist.keys();
                    while (doclist.hasMoreElements()) {
                        docid = (String) doclist.nextElement();
                        document = (String) nodelist.get(docid);
                        result.append("  <document>\n    " + document
                                + "\n  </document>\n");
                    }

                    result.append("</resultset>\n");
                }
                // Time the request if asked for
                double stopTime = System.currentTimeMillis();
                double dbOpenTime = (connTime - startTime) / 1000;
                double readTime = (stopTime - connTime) / 1000;
                double executionTime = (stopTime - startTime) / 1000;
                if (showRuntime) {
                    System.out.print("  " + executionTime);
                    System.out.print("  " + dbOpenTime);
                    System.out.print("  " + readTime);
                    System.out.print("  " + nodelist.size());
                    System.out.println();
                }
                //System.out.println(result);
                //write into a file "result.txt"
                if (!showRuntime) {
                    File f = new File("./result.txt");
                    FileWriter fw = new FileWriter(f);
                    BufferedWriter out = new BufferedWriter(fw);
                    out.write(result.toString());
                    out.flush();
                    out.close();
                    fw.close();
                }

            } catch (Exception e) {
                System.err.println("Error in DBQuery.main");
                System.err.println(e.getMessage());
                e.printStackTrace(System.err);
            }
        }
    }

    /**
     * construct an instance of the DBQuery class
     *
     * <p>
     * Generally, one would call the findDocuments() routine after creating an
     * instance to specify the search query
     * </p>
     *

     * @param parserName the fully qualified name of a Java class implementing
     *            the org.xml.sax.XMLReader interface
     */
    public DBQuery(String parserName)
    {
       this.parserName = parserName;
    }


  /**
   * Method put the search result set into out printerwriter
   * @param resoponse the return response
   * @param out the output printer
   * @param params the paratermer hashtable
   * @param user the user name (it maybe different to the one in param)
   * @param groups the group array
   * @param sessionid  the sessionid
   */
  public void findDocuments(HttpServletResponse response,
                                       PrintWriter out, Hashtable params,
                                       String user, String[] groups,
                                       String sessionid)
  {
    boolean useXMLIndex = (new Boolean(MetaCatUtil.getOption("usexmlindex")))
               .booleanValue();
    findDocuments(response, out, params, user, groups, sessionid, useXMLIndex);

  }


    /**
     * Method put the search result set into out printerwriter
     * @param resoponse the return response
     * @param out the output printer
     * @param params the paratermer hashtable
     * @param user the user name (it maybe different to the one in param)
     * @param groups the group array
     * @param sessionid  the sessionid
     */
    public void findDocuments(HttpServletResponse response,
                                         PrintWriter out, Hashtable params,
                                         String user, String[] groups,
                                         String sessionid, boolean useXMLIndex)
    {
      // get query and qformat
      String xmlquery = ((String[])params.get("query"))[0];

      MetaCatUtil.debugMessage("xmlquery: " + xmlquery, 30);
      String qformat = ((String[])params.get("qformat"))[0];
      MetaCatUtil.debugMessage("qformat: " + qformat, 30);
      // Get the XML query and covert it into a SQL statment
      QuerySpecification qspec = null;
      if ( xmlquery != null)
      {
         xmlquery = transformQuery(xmlquery);
         try
         {
           qspec = new QuerySpecification(xmlquery,
                                          parserName,
                                          MetaCatUtil.getOption("accNumSeparator"));
         }
         catch (Exception ee)
         {
           MetaCatUtil.debugMessage("erro to generate QuerySpecification object"
                                    +" in DBQuery.findDocuments"
                                    + ee.getMessage(), 30);
         }
      }



      if (qformat != null && qformat.equals(MetaCatServlet.XMLFORMAT))
      {
        //xml format
        response.setContentType("text/xml");
        createResultDocument(xmlquery, qspec, out, user, groups, useXMLIndex);
      }//if
      else
      {
        //knb format, in this case we will get whole result and sent it out
        response.setContentType("text/html");
        PrintWriter nonout = null;
        StringBuffer xml = createResultDocument(xmlquery, qspec, nonout, user,
                                                groups, useXMLIndex);
        //transfer the xml to html
        try
        {

         DBTransform trans = new DBTransform();
         response.setContentType("text/html");
         trans.transformXMLDocument(xml.toString(), "-//NCEAS//resultset//EN",
                                 "-//W3C//HTML//EN", qformat, out, params,
                                 sessionid);

        }
        catch(Exception e)
        {
         MetaCatUtil.debugMessage("Error in MetaCatServlet.transformResultset:"
                                +e.getMessage(), 30);
         }

      }//else

    }

  /*
   * Transforms a hashtable of documents to an xml or html result and sent
   * the content to outputstream. Keep going untill hastable is empty. stop it.
   * add the QuerySpecification as parameter is for ecogrid. But it is duplicate
   * to xmlquery String
   */
  public StringBuffer createResultDocument(String xmlquery,
                                            QuerySpecification qspec,
                                            PrintWriter out,
                                            String user, String[] groups,
                                            boolean useXMLIndex)
  {
    DBConnection dbconn = null;
    int serialNumber = -1;
    StringBuffer resultset = new StringBuffer();
    resultset.append("<?xml version=\"1.0\"?>\n");
    resultset.append("<resultset>\n");
    resultset.append("  <query>" + xmlquery + "</query>");
    // sent query part out
    if (out != null)
    {
      out.println(resultset.toString());
    }
    if (qspec != null)
    {
      try
      {

        //checkout the dbconnection
        dbconn = DBConnectionPool.getDBConnection("DBQuery.findDocuments");
        serialNumber = dbconn.getCheckOutSerialNumber();

        //print out the search result
        // search the doc list
        resultset = findResultDoclist(qspec, resultset, out, user, groups,
                                      dbconn, useXMLIndex);

      } //try
      catch (IOException ioe)
      {
        MetaCatUtil.debugMessage("IO error in DBQuery.findDocuments:", 30);
        MetaCatUtil.debugMessage(ioe.getMessage(), 30);

      }
      catch (SQLException e)
      {
        MetaCatUtil.debugMessage("SQL Error in DBQuery.findDocuments: "
                                 + e.getMessage(), 30);
      }
      catch (Exception ee)
      {
        MetaCatUtil.debugMessage("Exception in DBQuery.findDocuments: "
                                 + ee.getMessage(), 30);
      }
      finally
      {
        DBConnectionPool.returnDBConnection(dbconn, serialNumber);
      } //finally
    }//if
    String closeRestultset = "</resultset>";
    resultset.append(closeRestultset);
    if (out != null)
    {
      out.println(closeRestultset);
    }

    return resultset;
  }//createResultDocuments



    /*
     * Find the doc list which match the query
     */
    private StringBuffer findResultDoclist(QuerySpecification qspec,
                                      StringBuffer resultsetBuffer,
                                      PrintWriter out,
                                      String user, String[]groups,
                                      DBConnection dbconn, boolean useXMLIndex )
                                      throws Exception
    {
      int offset = 1;
      // this is a hack for offset
      if (out == null)
      {
        // for html page, we put everything into one page
        offset =
            (new Integer(MetaCatUtil.getOption("web_resultsetsize"))).intValue();
      }
      else
      {
          offset =
              (new Integer(MetaCatUtil.getOption("app_resultsetsize"))).intValue();
      }

      int count = 0;
      int index = 0;
      Hashtable docListResult = new Hashtable();
      PreparedStatement pstmt = null;
      String docid = null;
      String docname = null;
      String doctype = null;
      String createDate = null;
      String updateDate = null;
      StringBuffer document = null;
      int rev = 0;
      String query = qspec.printSQL(useXMLIndex);
      String ownerQuery = getOwnerQuery(user);
      MetaCatUtil.debugMessage("query: " + query, 30);
      //MetaCatUtil.debugMessage("query: "+ownerQuery, 30);
      // if query is not the owner query, we need to check the permission
      // otherwise we don't need (owner has all permission by default)
      if (!query.equals(ownerQuery))
      {
        // set user name and group
        qspec.setUserName(user);
        qspec.setGroup(groups);
        // Get access query
        String accessQuery = qspec.getAccessQuery();
        if(!query.endsWith("WHERE")){
            query = query + accessQuery;
        } else {
            query = query + accessQuery.substring(4, accessQuery.length());
        }
        MetaCatUtil.debugMessage(" final query: " + query, 30);
      }

      double startTime = System.currentTimeMillis() / 1000;
      pstmt = dbconn.prepareStatement(query);

      // Execute the SQL query using the JDBC connection
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      double queryExecuteTime = System.currentTimeMillis() / 1000;
      MetaCatUtil.debugMessage("Time for execute query: "
                    + (queryExecuteTime - startTime), 30);
      boolean tableHasRows = rs.next();
      while (tableHasRows)
      {
        docid = rs.getString(1).trim();
        docname = rs.getString(2);
        doctype = rs.getString(3);
        createDate = rs.getString(4);
        updateDate = rs.getString(5);
        rev = rs.getInt(6);

        // if there are returndocs to match, backtracking can be performed
        // otherwise, just return the document that was hit
        Vector returndocVec = qspec.getReturnDocList();
         if (returndocVec.size() != 0 && !returndocVec.contains(doctype)
                        && !qspec.isPercentageSearch())
        {
           MetaCatUtil.debugMessage("Back tracing now...", 20);
           String sep = MetaCatUtil.getOption("accNumSeparator");
           StringBuffer btBuf = new StringBuffer();
           btBuf.append("select docid from xml_relation where ");

           //build the doctype list for the backtracking sql statement
           btBuf.append("packagetype in (");
           for (int i = 0; i < returndocVec.size(); i++)
           {
             btBuf.append("'").append((String) returndocVec.get(i)).append("'");
             if (i != (returndocVec.size() - 1))
             {
                btBuf.append(", ");
              }
            }
            btBuf.append(") ");
            btBuf.append("and (subject like '");
            btBuf.append(docid).append("'");
            btBuf.append("or object like '");
            btBuf.append(docid).append("')");

            PreparedStatement npstmt = dbconn.prepareStatement(btBuf.toString());
            //should incease usage count
            dbconn.increaseUsageCount(1);
            npstmt.execute();
            ResultSet btrs = npstmt.getResultSet();
            boolean hasBtRows = btrs.next();
            while (hasBtRows)
            {
               //there was a backtrackable document found
               DocumentImpl xmldoc = null;
               String packageDocid = btrs.getString(1);
               MetaCatUtil.debugMessage("Getting document for docid: "
                                         + packageDocid, 40);
                try
                {
                    //  THIS CONSTRUCTOR BUILDS THE WHOLE XML doc not
                    // needed here
                    // xmldoc = new DocumentImpl(dbconn, packageDocid);
                    //  thus use the following to get the doc info only
                    //  xmldoc = new DocumentImpl(dbconn);
                    xmldoc = new DocumentImpl(packageDocid, false);
                    if (xmldoc == null)
                    {
                       MetaCatUtil.debugMessage("Document was null for: "
                                                + packageDocid, 50);
                    }
                }
                catch (Exception e)
                {
                    System.out.println("Error getting document in "
                                       + "DBQuery.findDocuments: "
                                       + e.getMessage());
                }

                String docid_org = xmldoc.getDocID();
                if (docid_org == null)
                {
                   MetaCatUtil.debugMessage("Docid_org was null.", 40);
                   //continue;
                }
                docid = docid_org.trim();
                docname = xmldoc.getDocname();
                doctype = xmldoc.getDoctype();
                createDate = xmldoc.getCreateDate();
                updateDate = xmldoc.getUpdateDate();
                rev = xmldoc.getRev();
                document = new StringBuffer();

                String completeDocid = docid
                                + MetaCatUtil.getOption("accNumSeparator");
                completeDocid += rev;
                document.append("<docid>").append(completeDocid);
                document.append("</docid>");
                if (docname != null)
                {
                  document.append("<docname>" + docname + "</docname>");
                }
                if (doctype != null)
                {
                  document.append("<doctype>" + doctype + "</doctype>");
                }
                if (createDate != null)
                {
                 document.append("<createdate>" + createDate + "</createdate>");
                }
                if (updateDate != null)
                {
                  document.append("<updatedate>" + updateDate+ "</updatedate>");
                }
                // Store the document id and the root node id
                docListResult.put(docid, (String) document.toString());
                count++;


                // Get the next package document linked to our hit
                hasBtRows = btrs.next();
              }//while
              npstmt.close();
              btrs.close();
        }
        else if (returndocVec.size() == 0 || returndocVec.contains(doctype))
        {

           document = new StringBuffer();

           String completeDocid = docid
                            + MetaCatUtil.getOption("accNumSeparator");
           completeDocid += rev;
           document.append("<docid>").append(completeDocid).append("</docid>");
           if (docname != null)
           {
               document.append("<docname>" + docname + "</docname>");
            }
            if (doctype != null)
            {
               document.append("<doctype>" + doctype + "</doctype>");
            }
            if (createDate != null)
            {
                document.append("<createdate>" + createDate + "</createdate>");
             }
             if (updateDate != null)
             {
               document.append("<updatedate>" + updateDate + "</updatedate>");
             }
              // Store the document id and the root node id
              docListResult.put(docid, (String) document.toString());
              count++;


        }//else
        // when doclist reached the offset number, send out doc list and empty
        // the hash table
        if (count == offset)
        {
          //reset count
          count = 0;
          handleSubsetResult(qspec,resultsetBuffer, out, docListResult,
                              user, groups,dbconn, useXMLIndex);
          // reset docListResult
          docListResult = new Hashtable();

        }
       // Advance to the next record in the cursor
       tableHasRows = rs.next();
     }//while
     rs.close();
     pstmt.close();
     //if docListResult is not empty, it need to be sent.
     if (!docListResult.isEmpty())
     {
       handleSubsetResult(qspec,resultsetBuffer, out, docListResult,
                              user, groups,dbconn, useXMLIndex);
     }
     double docListTime = System.currentTimeMillis() / 1000;
     MetaCatUtil.debugMessage("prepare docid list time: "
                    + (docListTime - queryExecuteTime), 30);

     return resultsetBuffer;
    }//findReturnDoclist


    /*
     * Send completed search hashtable(part of reulst)to output stream
     * and buffer into a buffer stream
     */
    private StringBuffer handleSubsetResult(QuerySpecification qspec,
                                           StringBuffer resultset,
                                           PrintWriter out, Hashtable partOfDoclist,
                                           String user, String[]groups,
                                       DBConnection dbconn, boolean useXMLIndex)
                                       throws Exception
   {

     // check if there is a record in xml_returnfield
     // and get the returnfield_id and usage count
     int usage_count = getXmlReturnfieldsTableId(qspec, dbconn);
     boolean enterRecords = false;

     // get value of xml_returnfield_count
     int count = (new Integer(MetaCatUtil
                            .getOption("xml_returnfield_count")))
                            .intValue();

     // set enterRecords to true if usage_count is more than the offset
     // specified in metacat.properties
     if(usage_count > count){
         enterRecords = true;
     }

     if(returnfield_id < 0){
         MetaCatUtil.debugMessage("Error in getting returnfield id from"
                                  + "xml_returnfield table", 20);
	enterRecords = false;
     }

     // get the hashtable containing the docids that already in the
     // xml_queryresult table
     MetaCatUtil.debugMessage("size of partOfDoclist before"
                             + " docidsInQueryresultTable(): "
                             + partOfDoclist.size() , 50);
     Hashtable queryresultDocList = docidsInQueryresultTable(returnfield_id,
                                                        partOfDoclist, dbconn);

     // remove the keys in queryresultDocList from partOfDoclist
     Enumeration _keys = queryresultDocList.keys();
     while (_keys.hasMoreElements()){
         partOfDoclist.remove(_keys.nextElement());
     }

     // backup the keys-elements in partOfDoclist to check later
     // if the doc entry is indexed yet
     Hashtable partOfDoclistBackup = new Hashtable();
     _keys = partOfDoclist.keys();
     while (_keys.hasMoreElements()){
	 Object key = _keys.nextElement();
         partOfDoclistBackup.put(key, partOfDoclist.get(key));
     }

     MetaCatUtil.debugMessage("size of partOfDoclist after"
                             + " docidsInQueryresultTable(): "
                             + partOfDoclist.size() , 50);

     //add return fields for the documents in partOfDoclist
     partOfDoclist = addReturnfield(partOfDoclist, qspec, user, groups,
                                        dbconn, useXMLIndex );
     //add relationship part part docid list for the documents in partOfDocList
     partOfDoclist = addRelationship(partOfDoclist, qspec, dbconn, useXMLIndex);


     Enumeration keys = partOfDoclist.keys();
     String key = null;
     String element = null;
     String query = null;
     int offset = (new Integer(MetaCatUtil
                               .getOption("queryresult_string_length")))
                               .intValue();
     while (keys.hasMoreElements())
     {
         key = (String) keys.nextElement();
         element = (String)partOfDoclist.get(key);

	 // check if the enterRecords is true, elements is not null, element's
         // length is less than the limit of table column and if the document
         // has been indexed already
         if(enterRecords && element != null
		&& element.length() < offset
		&& element.compareTo((String) partOfDoclistBackup.get(key)) != 0){
             query = "INSERT INTO xml_queryresult (returnfield_id, docid, "
                 + "queryresult_string) VALUES (?, ?, ?)";

             PreparedStatement pstmt = null;
             pstmt = dbconn.prepareStatement(query);
             pstmt.setInt(1, returnfield_id);
             pstmt.setString(2, key);
             pstmt.setString(3, element);

             dbconn.increaseUsageCount(1);
             pstmt.execute();
             pstmt.close();
         }

         // A string with element
         String xmlElement = "  <document>" + element + "</document>";

         //send single element to output
         if (out != null)
         {
             out.println(xmlElement);
         }
         resultset.append(xmlElement);
     }//while


     keys = queryresultDocList.keys();
     while (keys.hasMoreElements())
     {
         key = (String) keys.nextElement();
         element = (String)queryresultDocList.get(key);
         // A string with element
         String xmlElement = "  <document>" + element + "</document>";
         //send single element to output
         if (out != null)
         {
             out.println(xmlElement);
         }
         resultset.append(xmlElement);
     }//while

     return resultset;
 }

   /**
    * Get the docids already in xml_queryresult table and corresponding
    * queryresultstring as a hashtable
    */
   private Hashtable docidsInQueryresultTable(int returnfield_id,
                                              Hashtable partOfDoclist,
                                              DBConnection dbconn){

         Hashtable returnValue = new Hashtable();
         PreparedStatement pstmt = null;
         ResultSet rs = null;

         // get partOfDoclist as string for the query
         Enumeration keylist = partOfDoclist.keys();
         StringBuffer doclist = new StringBuffer();
         while (keylist.hasMoreElements())
         {
             doclist.append("'");
             doclist.append((String) keylist.nextElement());
             doclist.append("',");
         }//while


         if (doclist.length() > 0)
         {
             doclist.deleteCharAt(doclist.length() - 1); //remove the last comma

             // the query to find out docids from xml_queryresult
             String query = "select docid, queryresult_string from "
                          + "xml_queryresult where returnfield_id = " +
                          returnfield_id +" and docid in ("+ doclist + ")";
             MetaCatUtil.debugMessage("Query to get docids from xml_queryresult:"
                                      + query, 50);

             try {
                 // prepare and execute the query
                 pstmt = dbconn.prepareStatement(query);
                 dbconn.increaseUsageCount(1);
                 pstmt.execute();
                 rs = pstmt.getResultSet();
                 boolean tableHasRows = rs.next();
                 while (tableHasRows) {
                     // store the returned results in the returnValue hashtable
                     String key = rs.getString(1);
                     String element = rs.getString(2);

                     if(element != null){
                         returnValue.put(key, element);
                     } else {
                         MetaCatUtil.debugMessage("Null elment found ("
                         + "DBQuery.docidsInQueryresultTable)", 50);
                     }
                     tableHasRows = rs.next();
                 }
                 rs.close();
                 pstmt.close();
             } catch (Exception e){
                 MetaCatUtil.debugMessage("Error getting docids from "
                                          + "queryresult in "
                                          + "DBQuery.docidsInQueryresultTable: "
                                          + e.getMessage(), 20);
              }
         }
         return returnValue;
     }


   /**
    * Method to get id from xml_returnfield table
    * for a given query specification
    */
   private int returnfield_id;
   private int getXmlReturnfieldsTableId(QuerySpecification qspec,
                                           DBConnection dbconn){
       int id = -1;
       int count = 1;
       PreparedStatement pstmt = null;
       ResultSet rs = null;
       String returnfield = qspec.getSortedReturnFieldString();

       // query for finding the id from xml_returnfield
       String query = "SELECT returnfield_id, usage_count FROM xml_returnfield "
            + "WHERE returnfield_string LIKE ?";
       MetaCatUtil.debugMessage("ReturnField Query:" + query, 50);

       try {
           // prepare and run the query
           pstmt = dbconn.prepareStatement(query);
           pstmt.setString(1,returnfield);
           dbconn.increaseUsageCount(1);
           pstmt.execute();
           rs = pstmt.getResultSet();
           boolean tableHasRows = rs.next();

           // if record found then increase the usage count
           // else insert a new record and get the id of the new record
           if(tableHasRows){
               // get the id
               id = rs.getInt(1);
               count = rs.getInt(2) + 1;
               rs.close();
               pstmt.close();

               // increase the usage count
               query = "UPDATE xml_returnfield SET usage_count ='" + count
                   + "' WHERE returnfield_id ='"+ id +"'";
               MetaCatUtil.debugMessage("ReturnField Table Update:"+ query, 50);

               pstmt = dbconn.prepareStatement(query);
               dbconn.increaseUsageCount(1);
               pstmt.execute();
               pstmt.close();

           } else {
               rs.close();
               pstmt.close();

               // insert a new record
               query = "INSERT INTO xml_returnfield (returnfield_string, usage_count)"
                   + "VALUES (?, '1')";
               MetaCatUtil.debugMessage("ReturnField Table Insert:"+ query, 50);
               pstmt = dbconn.prepareStatement(query);
               pstmt.setString(1, returnfield);
               dbconn.increaseUsageCount(1);
               pstmt.execute();
               pstmt.close();

               // get the id of the new record
               query = "SELECT returnfield_id FROM xml_returnfield "
                   + "WHERE returnfield_string LIKE ?";
               MetaCatUtil.debugMessage("ReturnField query after Insert:" + query, 50);
               pstmt = dbconn.prepareStatement(query);
               pstmt.setString(1, returnfield);

               dbconn.increaseUsageCount(1);
               pstmt.execute();
               rs = pstmt.getResultSet();
               if(rs.next()){
                   id = rs.getInt(1);
               } else {
                   id = -1;
               }
               rs.close();
               pstmt.close();
           }

       } catch (Exception e){
           MetaCatUtil.debugMessage("Error getting id from xml_returnfield in "
                                     + "DBQuery.getXmlReturnfieldsTableId: "
                                     + e.getMessage(), 20);
           id = -1;
       }

       returnfield_id = id;
       return count;
   }


    /*
     * A method to add return field to return doclist hash table
     */
    private Hashtable addReturnfield(Hashtable docListResult,
                                      QuerySpecification qspec,
                                      String user, String[]groups,
                                      DBConnection dbconn, boolean useXMLIndex )
                                      throws Exception
    {
      PreparedStatement pstmt = null;
      ResultSet rs = null;
      String docid = null;
      String fieldname = null;
      String fielddata = null;
      String relation = null;

      if (qspec.containsExtendedSQL())
      {
        qspec.setUserName(user);
        qspec.setGroup(groups);
        Vector extendedFields = new Vector(qspec.getReturnFieldList());
        Vector results = new Vector();
        Enumeration keylist = docListResult.keys();
        StringBuffer doclist = new StringBuffer();
        Vector parentidList = new Vector();
        Hashtable returnFieldValue = new Hashtable();
        while (keylist.hasMoreElements())
        {
          doclist.append("'");
          doclist.append((String) keylist.nextElement());
          doclist.append("',");
        }
        if (doclist.length() > 0)
        {
          Hashtable controlPairs = new Hashtable();
          double extendedQueryStart = System.currentTimeMillis() / 1000;
          doclist.deleteCharAt(doclist.length() - 1); //remove the last comma
          // check if user has permission to see the return field data
          String accessControlSQL =
                 qspec.printAccessControlSQLForReturnField(doclist.toString());
          pstmt = dbconn.prepareStatement(accessControlSQL);
          //increase dbconnection usage count
          dbconn.increaseUsageCount(1);
          pstmt.execute();
          rs = pstmt.getResultSet();
          boolean tableHasRows = rs.next();
          while (tableHasRows)
          {
            long startNodeId = rs.getLong(1);
            long endNodeId = rs.getLong(2);
            controlPairs.put(new Long(startNodeId), new Long(endNodeId));
            tableHasRows = rs.next();
          }

           double extendedAccessQueryEnd = System.currentTimeMillis() / 1000;
           MetaCatUtil.debugMessage( "Time for execute access extended query: "
                          + (extendedAccessQueryEnd - extendedQueryStart),30);

           String extendedQuery =
               qspec.printExtendedSQL(doclist.toString(), controlPairs, useXMLIndex);
           MetaCatUtil.debugMessage("Extended query: " + extendedQuery, 30);

           if(extendedQuery != null){
               pstmt = dbconn.prepareStatement(extendedQuery);
               //increase dbconnection usage count
               dbconn.increaseUsageCount(1);
               pstmt.execute();
               rs = pstmt.getResultSet();
               double extendedQueryEnd = System.currentTimeMillis() / 1000;
               MetaCatUtil.debugMessage(
                   "Time for execute extended query: "
                   + (extendedQueryEnd - extendedQueryStart), 30);
               tableHasRows = rs.next();
               while (tableHasRows) {
                   ReturnFieldValue returnValue = new ReturnFieldValue();
                   docid = rs.getString(1).trim();
                   fieldname = rs.getString(2);
                   fielddata = rs.getString(3);
                   fielddata = MetaCatUtil.normalize(fielddata);
                   String parentId = rs.getString(4);
                   StringBuffer value = new StringBuffer();

                   // if xml_index is used, there would be just one record per nodeid
                   // as xml_index just keeps one entry for each path
                   if (useXMLIndex || !containsKey(parentidList, parentId)) {
                       // don't need to merger nodedata
                       value.append("<param name=\"");
                       value.append(fieldname);
                       value.append("\">");
                       value.append(fielddata);
                       value.append("</param>");
                       //set returnvalue
                       returnValue.setDocid(docid);
                       returnValue.setFieldValue(fielddata);
                       returnValue.setXMLFieldValue(value.toString());
                       // Store it in hastable
                       putInArray(parentidList, parentId, returnValue);
                   }
                   else {
                       // need to merge nodedata if they have same parent id and
                       // node type is text
                       fielddata = (String) ( (ReturnFieldValue)
                                             getArrayValue(
                           parentidList, parentId)).getFieldValue()
                           + fielddata;
                       value.append("<param name=\"");
                       value.append(fieldname);
                       value.append("\">");
                       value.append(fielddata);
                       value.append("</param>");
                       returnValue.setDocid(docid);
                       returnValue.setFieldValue(fielddata);
                       returnValue.setXMLFieldValue(value.toString());
                       // remove the old return value from paretnidList
                       parentidList.remove(parentId);
                       // store the new return value in parentidlit
                       putInArray(parentidList, parentId, returnValue);
                   }
                   tableHasRows = rs.next();
               } //while
               rs.close();
               pstmt.close();

               // put the merger node data info into doclistReult
               Enumeration xmlFieldValue = (getElements(parentidList)).
                   elements();
               while (xmlFieldValue.hasMoreElements()) {
                   ReturnFieldValue object =
                       (ReturnFieldValue) xmlFieldValue.nextElement();
                   docid = object.getDocid();
                   if (docListResult.containsKey(docid)) {
                       String removedelement = (String) docListResult.
                           remove(docid);
                       docListResult.
                           put(docid,
                               removedelement + object.getXMLFieldValue());
                   }
                   else {
                       docListResult.put(docid, object.getXMLFieldValue());
                   }
               } //while
               double docListResultEnd = System.currentTimeMillis() / 1000;
               MetaCatUtil.debugMessage(
                   "Time for prepare doclistresult after"
                   + " execute extended query: "
                   + (docListResultEnd - extendedQueryEnd),
                   30);
           }

           // get attribures return
           docListResult = getAttributeValueForReturn(qspec,
                           docListResult, doclist.toString(), useXMLIndex);
       }//if doclist lenght is great than zero

     }//if has extended query

      return docListResult;
    }//addReturnfield

    /*
    * A method to add relationship to return doclist hash table
    */
   private Hashtable addRelationship(Hashtable docListResult,
                                     QuerySpecification qspec,
                                     DBConnection dbconn, boolean useXMLIndex )
                                     throws Exception
  {
    PreparedStatement pstmt = null;
    ResultSet rs = null;
    StringBuffer document = null;
    double startRelation = System.currentTimeMillis() / 1000;
    Enumeration docidkeys = docListResult.keys();
    while (docidkeys.hasMoreElements())
    {
      //String connstring =
      // "metacat://"+util.getOption("server")+"?docid=";
      String connstring = "%docid=";
      String docidkey = (String) docidkeys.nextElement();
      pstmt = dbconn.prepareStatement(QuerySpecification
                      .printRelationSQL(docidkey));
      pstmt.execute();
      rs = pstmt.getResultSet();
      boolean tableHasRows = rs.next();
      while (tableHasRows)
      {
        String sub = rs.getString(1);
        String rel = rs.getString(2);
        String obj = rs.getString(3);
        String subDT = rs.getString(4);
        String objDT = rs.getString(5);

        document = new StringBuffer();
        document.append("<triple>");
        document.append("<subject>").append(MetaCatUtil.normalize(sub));
        document.append("</subject>");
        if (subDT != null)
        {
          document.append("<subjectdoctype>").append(subDT);
          document.append("</subjectdoctype>");
        }
        document.append("<relationship>").append(MetaCatUtil.normalize(rel));
        document.append("</relationship>");
        document.append("<object>").append(MetaCatUtil.normalize(obj));
        document.append("</object>");
        if (objDT != null)
        {
          document.append("<objectdoctype>").append(objDT);
          document.append("</objectdoctype>");
        }
        document.append("</triple>");

        String removedelement = (String) docListResult.remove(docidkey);
        docListResult.put(docidkey, removedelement+ document.toString());
        tableHasRows = rs.next();
      }//while
      rs.close();
      pstmt.close();
    }//while
    double endRelation = System.currentTimeMillis() / 1000;
    MetaCatUtil.debugMessage("Time for adding relation to docListResult: "
                             + (endRelation - startRelation), 30);

    return docListResult;
  }//addRelation

  /**
   * removes the <?xml version="1.0"?> tag from the beginning.  This takes a
   * string as a param instead of a hashtable.
   *
   * @param xmlquery a string representing a query.
   */
   private  String transformQuery(String xmlquery)
   {
     xmlquery = xmlquery.trim();
     int index = xmlquery.indexOf("?>");
     if (index != -1)
     {
       return xmlquery.substring(index + 2, xmlquery.length());
     }
     else
     {
       return xmlquery;
     }
   }


    /*
     * A method to search if Vector contains a particular key string
     */
    private boolean containsKey(Vector parentidList, String parentId)
    {

        Vector tempVector = null;

        for (int count = 0; count < parentidList.size(); count++) {
            tempVector = (Vector) parentidList.get(count);
            if (parentId.compareTo((String) tempVector.get(0)) == 0) { return true; }
        }
        return false;
    }

    /*
     * A method to put key and value in Vector
     */
    private void putInArray(Vector parentidList, String key,
            ReturnFieldValue value)
    {

        Vector tempVector = null;

        for (int count = 0; count < parentidList.size(); count++) {
            tempVector = (Vector) parentidList.get(count);

            if (key.compareTo((String) tempVector.get(0)) == 0) {
                tempVector.remove(1);
                tempVector.add(1, value);
                return;
            }
        }

        tempVector = new Vector();
        tempVector.add(0, key);
        tempVector.add(1, value);
        parentidList.add(tempVector);
        return;
    }

    /*
     * A method to get value in Vector given a key
     */
    private ReturnFieldValue getArrayValue(Vector parentidList, String key)
    {

        Vector tempVector = null;

        for (int count = 0; count < parentidList.size(); count++) {
            tempVector = (Vector) parentidList.get(count);

            if (key.compareTo((String) tempVector.get(0)) == 0) { return (ReturnFieldValue) tempVector
                    .get(1); }
        }
        return null;
    }

    /*
     * A method to get enumeration of all values in Vector
     */
    private Vector getElements(Vector parentidList)
    {
        Vector enumVector = new Vector();
        Vector tempVector = null;

        for (int count = 0; count < parentidList.size(); count++) {
            tempVector = (Vector) parentidList.get(count);

            enumVector.add(tempVector.get(1));
        }
        return enumVector;
    }

    /*
     * A method to return search result after running a query which return
     * field have attribue
     */
    private Hashtable getAttributeValueForReturn(QuerySpecification squery,
            Hashtable docInformationList, String docList, boolean useXMLIndex)
    {
        StringBuffer XML = null;
        String sql = null;
        DBConnection dbconn = null;
        PreparedStatement pstmt = null;
        ResultSet rs = null;
        int serialNumber = -1;
        boolean tableHasRows = false;

        //check the parameter
        if (squery == null || docList == null || docList.length() < 0) { return docInformationList; }

        // if has attribute as return field
        if (squery.containsAttributeReturnField()) {
            sql = squery.printAttributeQuery(docList, useXMLIndex);
            try {
                dbconn = DBConnectionPool
                        .getDBConnection("DBQuery.getAttributeValue");
                serialNumber = dbconn.getCheckOutSerialNumber();
                pstmt = dbconn.prepareStatement(sql);
                pstmt.execute();
                rs = pstmt.getResultSet();
                tableHasRows = rs.next();
                while (tableHasRows) {
                    String docid = rs.getString(1).trim();
                    String fieldname = rs.getString(2);
                    String fielddata = rs.getString(3);
                    String attirbuteName = rs.getString(4);
                    XML = new StringBuffer();

                    XML.append("<param name=\"");
                    XML.append(fieldname);
                    XML.append("/");
                    XML.append(QuerySpecification.ATTRIBUTESYMBOL);
                    XML.append(attirbuteName);
                    XML.append("\">");
                    XML.append(fielddata);
                    XML.append("</param>");
                    tableHasRows = rs.next();

                    if (docInformationList.containsKey(docid)) {
                        String removedelement = (String) docInformationList
                                .remove(docid);
                        docInformationList.put(docid, removedelement
                                + XML.toString());
                    } else {
                        docInformationList.put(docid, XML.toString());
                    }
                }//while
                rs.close();
                pstmt.close();
            } catch (Exception se) {
                MetaCatUtil.debugMessage(
                        "Error in DBQuery.getAttributeValue1: "
                                + se.getMessage(), 30);
            } finally {
                try {
                    pstmt.close();
                }//try
                catch (SQLException sqlE) {
                    MetaCatUtil.debugMessage(
                            "Error in DBQuery.getAttributeValue2: "
                                    + sqlE.getMessage(), 30);
                }//catch
                finally {
                    DBConnectionPool.returnDBConnection(dbconn, serialNumber);
                }//finally
            }//finally
        }//if
        return docInformationList;

    }

    /*
     * A method to create a query to get owner's docid list
     */
    private String getOwnerQuery(String owner)
    {
        if (owner != null) {
            owner = owner.toLowerCase();
        }
        StringBuffer self = new StringBuffer();

        self.append("SELECT docid,docname,doctype,");
        self.append("date_created, date_updated, rev ");
        self.append("FROM xml_documents WHERE docid IN (");
        self.append("(");
        self.append("SELECT DISTINCT docid FROM xml_nodes WHERE \n");
        self.append("nodedata LIKE '%%%' ");
        self.append(") \n");
        self.append(") ");
        self.append(" AND (");
        self.append(" lower(user_owner) = '" + owner + "'");
        self.append(") ");
        return self.toString();
    }

    /**
     * format a structured query as an XML document that conforms to the
     * pathquery.dtd and is appropriate for submission to the DBQuery
     * structured query engine
     *
     * @param params The list of parameters that should be included in the
     *            query
     */
    public static String createSQuery(Hashtable params)
    {
        StringBuffer query = new StringBuffer();
        Enumeration elements;
        Enumeration keys;
        String filterDoctype = null;
        String casesensitive = null;
        String searchmode = null;
        Object nextkey;
        Object nextelement;
        //add the xml headers
        query.append("<?xml version=\"1.0\"?>\n");
        query.append("<pathquery version=\"1.2\">\n");



        if (params.containsKey("meta_file_id")) {
            query.append("<meta_file_id>");
            query.append(((String[]) params.get("meta_file_id"))[0]);
            query.append("</meta_file_id>");
        }

        if (params.containsKey("returndoctype")) {
            String[] returnDoctypes = ((String[]) params.get("returndoctype"));
            for (int i = 0; i < returnDoctypes.length; i++) {
                String doctype = (String) returnDoctypes[i];

                if (!doctype.equals("any") && !doctype.equals("ANY")
                        && !doctype.equals("")) {
                    query.append("<returndoctype>").append(doctype);
                    query.append("</returndoctype>");
                }
            }
        }

        if (params.containsKey("filterdoctype")) {
            String[] filterDoctypes = ((String[]) params.get("filterdoctype"));
            for (int i = 0; i < filterDoctypes.length; i++) {
                query.append("<filterdoctype>").append(filterDoctypes[i]);
                query.append("</filterdoctype>");
            }
        }

        if (params.containsKey("returnfield")) {
            String[] returnfield = ((String[]) params.get("returnfield"));
            for (int i = 0; i < returnfield.length; i++) {
                query.append("<returnfield>").append(returnfield[i]);
                query.append("</returnfield>");
            }
        }

        if (params.containsKey("owner")) {
            String[] owner = ((String[]) params.get("owner"));
            for (int i = 0; i < owner.length; i++) {
                query.append("<owner>").append(owner[i]);
                query.append("</owner>");
            }
        }

        if (params.containsKey("site")) {
            String[] site = ((String[]) params.get("site"));
            for (int i = 0; i < site.length; i++) {
                query.append("<site>").append(site[i]);
                query.append("</site>");
            }
        }

        //allows the dynamic switching of boolean operators
        if (params.containsKey("operator")) {
            query.append("<querygroup operator=\""
                    + ((String[]) params.get("operator"))[0] + "\">");
        } else { //the default operator is UNION
            query.append("<querygroup operator=\"UNION\">");
        }

        if (params.containsKey("casesensitive")) {
            casesensitive = ((String[]) params.get("casesensitive"))[0];
        } else {
            casesensitive = "false";
        }

        if (params.containsKey("searchmode")) {
            searchmode = ((String[]) params.get("searchmode"))[0];
        } else {
            searchmode = "contains";
        }

        //anyfield is a special case because it does a
        //free text search. It does not have a <pathexpr>
        //tag. This allows for a free text search within the structured
        //query. This is useful if the INTERSECT operator is used.
        if (params.containsKey("anyfield")) {
            String[] anyfield = ((String[]) params.get("anyfield"));
            //allow for more than one value for anyfield
            for (int i = 0; i < anyfield.length; i++) {
                if (!anyfield[i].equals("")) {
                    query.append("<queryterm casesensitive=\"" + casesensitive
                            + "\" " + "searchmode=\"" + searchmode
                            + "\"><value>" + anyfield[i]
                            + "</value></queryterm>");
                }
            }
        }

        //this while loop finds the rest of the parameters
        //and attempts to query for the field specified
        //by the parameter.
        elements = params.elements();
        keys = params.keys();
        while (keys.hasMoreElements() && elements.hasMoreElements()) {
            nextkey = keys.nextElement();
            nextelement = elements.nextElement();

            //make sure we aren't querying for any of these
            //parameters since the are already in the query
            //in one form or another.
            Vector ignoredParams = new Vector();
            ignoredParams.add("returndoctype");
            ignoredParams.add("filterdoctype");
            ignoredParams.add("action");
            ignoredParams.add("qformat");
            ignoredParams.add("anyfield");
            ignoredParams.add("returnfield");
            ignoredParams.add("owner");
            ignoredParams.add("site");
            ignoredParams.add("operator");
            ignoredParams.add("sessionid");

            // Also ignore parameters listed in the properties file
            // so that they can be passed through to stylesheets
            String paramsToIgnore = MetaCatUtil
                    .getOption("query.ignored.params");
            StringTokenizer st = new StringTokenizer(paramsToIgnore, ",");
            while (st.hasMoreTokens()) {
                ignoredParams.add(st.nextToken());
            }
            if (!ignoredParams.contains(nextkey.toString())) {
                //allow for more than value per field name
                for (int i = 0; i < ((String[]) nextelement).length; i++) {
                    if (!((String[]) nextelement)[i].equals("")) {
                        query.append("<queryterm casesensitive=\""
                                + casesensitive + "\" " + "searchmode=\""
                                + searchmode + "\">" + "<value>" +
                                //add the query value
                                ((String[]) nextelement)[i]
                                + "</value><pathexpr>" +
                                //add the path to query by
                                nextkey.toString() + "</pathexpr></queryterm>");
                    }
                }
            }
        }
        query.append("</querygroup></pathquery>");
        //append on the end of the xml and return the result as a string
        return query.toString();
    }

    /**
     * format a simple free-text value query as an XML document that conforms
     * to the pathquery.dtd and is appropriate for submission to the DBQuery
     * structured query engine
     *
     * @param value the text string to search for in the xml catalog
     * @param doctype the type of documents to include in the result set -- use
     *            "any" or "ANY" for unfiltered result sets
     */
    public static String createQuery(String value, String doctype)
    {
        StringBuffer xmlquery = new StringBuffer();
        xmlquery.append("<?xml version=\"1.0\"?>\n");
        xmlquery.append("<pathquery version=\"1.0\">");

        if (!doctype.equals("any") && !doctype.equals("ANY")) {
            xmlquery.append("<returndoctype>");
            xmlquery.append(doctype).append("</returndoctype>");
        }

        xmlquery.append("<querygroup operator=\"UNION\">");
        //chad added - 8/14
        //the if statement allows a query to gracefully handle a null
        //query. Without this if a nullpointerException is thrown.
        if (!value.equals("")) {
            xmlquery.append("<queryterm casesensitive=\"false\" ");
            xmlquery.append("searchmode=\"contains\">");
            xmlquery.append("<value>").append(value).append("</value>");
            xmlquery.append("</queryterm>");
        }
        xmlquery.append("</querygroup>");
        xmlquery.append("</pathquery>");

        return (xmlquery.toString());
    }

    /**
     * format a simple free-text value query as an XML document that conforms
     * to the pathquery.dtd and is appropriate for submission to the DBQuery
     * structured query engine
     *
     * @param value the text string to search for in the xml catalog
     */
    public static String createQuery(String value)
    {
        return createQuery(value, "any");
    }

    /**
     * Check for "READ" permission on @docid for @user and/or @group from DB
     * connection
     */
    private boolean hasPermission(String user, String[] groups, String docid)
            throws SQLException, Exception
    {
        // Check for READ permission on @docid for @user and/or @groups
        PermissionController controller = new PermissionController(docid);
        return controller.hasPermission(user, groups,
                AccessControlInterface.READSTRING);
    }

    /**
     * Get all docIds list for a data packadge
     *
     * @param dataPackageDocid, the string in docId field of xml_relation table
     */
    private Vector getCurrentDocidListForDataPackage(String dataPackageDocid)
    {
        DBConnection dbConn = null;
        int serialNumber = -1;
        Vector docIdList = new Vector();//return value
        PreparedStatement pStmt = null;
        ResultSet rs = null;
        String docIdInSubjectField = null;
        String docIdInObjectField = null;

        // Check the parameter
        if (dataPackageDocid == null || dataPackageDocid.equals("")) { return docIdList; }//if

        //the query stirng
        String query = "SELECT subject, object from xml_relation where docId = ?";
        try {
            dbConn = DBConnectionPool
                    .getDBConnection("DBQuery.getCurrentDocidListForDataPackage");
            serialNumber = dbConn.getCheckOutSerialNumber();
            pStmt = dbConn.prepareStatement(query);
            //bind the value to query
            pStmt.setString(1, dataPackageDocid);

            //excute the query
            pStmt.execute();
            //get the result set
            rs = pStmt.getResultSet();
            //process the result
            while (rs.next()) {
                //In order to get the whole docIds in a data packadge,
                //we need to put the docIds of subject and object field in
                // xml_relation
                //into the return vector
                docIdInSubjectField = rs.getString(1);//the result docId in
                                                      // subject field
                docIdInObjectField = rs.getString(2);//the result docId in
                                                     // object field

                //don't put the duplicate docId into the vector
                if (!docIdList.contains(docIdInSubjectField)) {
                    docIdList.add(docIdInSubjectField);
                }

                //don't put the duplicate docId into the vector
                if (!docIdList.contains(docIdInObjectField)) {
                    docIdList.add(docIdInObjectField);
                }
            }//while
            //close the pStmt
            pStmt.close();
        }//try
        catch (SQLException e) {
            MetaCatUtil.debugMessage("Error in getDocidListForDataPackage: "
                    + e.getMessage(), 30);
        }//catch
        finally {
            try {
                pStmt.close();
            }//try
            catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "Error in getDocidListForDataPackage: "
                                + ee.getMessage(), 30);
            }//catch
            finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }//fianlly
        }//finally
        return docIdList;
    }//getCurrentDocidListForDataPackadge()

    /**
     * Get all docIds list for a data packadge
     *
     * @param dataPackageDocid, the string in docId field of xml_relation table
     */
    private Vector getOldVersionDocidListForDataPackage(String dataPackageDocid)
    {

        Vector docIdList = new Vector();//return value
        Vector tripleList = null;
        String xml = null;

        // Check the parameter
        if (dataPackageDocid == null || dataPackageDocid.equals("")) { return docIdList; }//if

        try {
            //initial a documentImpl object
            DocumentImpl packageDocument = new DocumentImpl(dataPackageDocid);
            //transfer to documentImpl object to string
            xml = packageDocument.toString();

            //create a tripcollection object
            TripleCollection tripleForPackage = new TripleCollection(
                    new StringReader(xml));
            //get the vetor of triples
            tripleList = tripleForPackage.getCollection();

            for (int i = 0; i < tripleList.size(); i++) {
                //put subject docid into docIdlist without duplicate
                if (!docIdList.contains(((Triple) tripleList.elementAt(i))
                        .getSubject())) {
                    //put subject docid into docIdlist
                    docIdList.add(((Triple) tripleList.get(i)).getSubject());
                }
                //put object docid into docIdlist without duplicate
                if (!docIdList.contains(((Triple) tripleList.elementAt(i))
                        .getObject())) {
                    docIdList.add(((Triple) (tripleList.get(i))).getObject());
                }
            }//for
        }//try
        catch (Exception e) {
            MetaCatUtil.debugMessage("Error in getOldVersionAllDocumentImpl: "
                    + e.getMessage(), 30);
        }//catch

        // return result
        return docIdList;
    }//getDocidListForPackageInXMLRevisions()

    /**
     * Check if the docId is a data packadge id. If the id is a data packadage
     * id, it should be store in the docId fields in xml_relation table. So we
     * can use a query to get the entries which the docId equals the given
     * value. If the result is null. The docId is not a packadge id. Otherwise,
     * it is.
     *
     * @param docId, the id need to be checked
     */
    private boolean isDataPackageId(String docId)
    {
        boolean result = false;
        PreparedStatement pStmt = null;
        ResultSet rs = null;
        String query = "SELECT docId from xml_relation where docId = ?";
        DBConnection dbConn = null;
        int serialNumber = -1;
        try {
            dbConn = DBConnectionPool
                    .getDBConnection("DBQuery.isDataPackageId");
            serialNumber = dbConn.getCheckOutSerialNumber();
            pStmt = dbConn.prepareStatement(query);
            //bind the value to query
            pStmt.setString(1, docId);
            //execute the query
            pStmt.execute();
            rs = pStmt.getResultSet();
            //process the result
            if (rs.next()) //There are some records for the id in docId fields
            {
                result = true;//It is a data packadge id
            }
            pStmt.close();
        }//try
        catch (SQLException e) {
            MetaCatUtil.debugMessage("Error in isDataPackageId: "
                    + e.getMessage(), 30);
        } finally {
            try {
                pStmt.close();
            }//try
            catch (SQLException ee) {
                MetaCatUtil.debugMessage("Error in isDataPackageId: "
                        + ee.getMessage(), 30);
            }//catch
            finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }//finally
        }//finally
        return result;
    }//isDataPackageId()

    /**
     * Check if the user has the permission to export data package
     *
     * @param conn, the connection
     * @param docId, the id need to be checked
     * @param user, the name of user
     * @param groups, the user's group
     */
    private boolean hasPermissionToExportPackage(String docId, String user,
            String[] groups) throws Exception
    {
        //DocumentImpl doc=new DocumentImpl(conn,docId);
        return DocumentImpl.hasReadPermission(user, groups, docId);
    }

    /**
     * Get the current Rev for a docid in xml_documents table
     *
     * @param docId, the id need to get version numb If the return value is -5,
     *            means no value in rev field for this docid
     */
    private int getCurrentRevFromXMLDoumentsTable(String docId)
            throws SQLException
    {
        int rev = -5;
        PreparedStatement pStmt = null;
        ResultSet rs = null;
        String query = "SELECT rev from xml_documents where docId = ?";
        DBConnection dbConn = null;
        int serialNumber = -1;
        try {
            dbConn = DBConnectionPool
                    .getDBConnection("DBQuery.getCurrentRevFromXMLDocumentsTable");
            serialNumber = dbConn.getCheckOutSerialNumber();
            pStmt = dbConn.prepareStatement(query);
            //bind the value to query
            pStmt.setString(1, docId);
            //execute the query
            pStmt.execute();
            rs = pStmt.getResultSet();
            //process the result
            if (rs.next()) //There are some records for rev
            {
                rev = rs.getInt(1);
                ;//It is the version for given docid
            } else {
                rev = -5;
            }

        }//try
        catch (SQLException e) {
            MetaCatUtil.debugMessage(
                    "Error in getCurrentRevFromXMLDoumentsTable: "
                            + e.getMessage(), 30);
            throw e;
        }//catch
        finally {
            try {
                pStmt.close();
            }//try
            catch (SQLException ee) {
                MetaCatUtil.debugMessage(
                        "Error in getCurrentRevFromXMLDoumentsTable: "
                                + ee.getMessage(), 30);
            }//catch
            finally {
                DBConnectionPool.returnDBConnection(dbConn, serialNumber);
            }//finally
        }//finally
        return rev;
    }//getCurrentRevFromXMLDoumentsTable

    /**
     * put a doc into a zip output stream
     *
     * @param docImpl, docmentImpl object which will be sent to zip output
     *            stream
     * @param zipOut, zip output stream which the docImpl will be put
     * @param packageZipEntry, the zip entry name for whole package
     */
    private void addDocToZipOutputStream(DocumentImpl docImpl,
            ZipOutputStream zipOut, String packageZipEntry)
            throws ClassNotFoundException, IOException, SQLException,
            McdbException, Exception
    {
        byte[] byteString = null;
        ZipEntry zEntry = null;

        byteString = docImpl.toString().getBytes();
        //use docId as the zip entry's name
        zEntry = new ZipEntry(packageZipEntry + "/metadata/"
                + docImpl.getDocID());
        zEntry.setSize(byteString.length);
        zipOut.putNextEntry(zEntry);
        zipOut.write(byteString, 0, byteString.length);
        zipOut.closeEntry();

    }//addDocToZipOutputStream()

    /**
     * Transfer a docid vetor to a documentImpl vector. The documentImpl vetor
     * only inlcudes current version. If a DocumentImple object couldn't find
     * for a docid, then the String of this docid was added to vetor rather
     * than DocumentImple object.
     *
     * @param docIdList, a vetor hold a docid list for a data package. In
     *            docid, there is not version number in it.
     */

    private Vector getCurrentAllDocumentImpl(Vector docIdList)
            throws McdbException, Exception
    {
        //Connection dbConn=null;
        Vector documentImplList = new Vector();
        int rev = 0;

        // Check the parameter
        if (docIdList.isEmpty()) { return documentImplList; }//if

        //for every docid in vector
        for (int i = 0; i < docIdList.size(); i++) {
            try {
                //get newest version for this docId
                rev = getCurrentRevFromXMLDoumentsTable((String) docIdList
                        .elementAt(i));

                // There is no record for this docId in xml_documents table
                if (rev == -5) {
                    // Rather than put DocumentImple object, put a String
                    // Object(docid)
                    // into the documentImplList
                    documentImplList.add((String) docIdList.elementAt(i));
                    // Skip other code
                    continue;
                }

                String docidPlusVersion = ((String) docIdList.elementAt(i))
                        + MetaCatUtil.getOption("accNumSeparator") + rev;

                //create new documentImpl object
                DocumentImpl documentImplObject = new DocumentImpl(
                        docidPlusVersion);
                //add them to vector
                documentImplList.add(documentImplObject);
            }//try
            catch (Exception e) {
                MetaCatUtil.debugMessage("Error in getCurrentAllDocumentImpl: "
                        + e.getMessage(), 30);
                // continue the for loop
                continue;
            }
        }//for
        return documentImplList;
    }

    /**
     * Transfer a docid vetor to a documentImpl vector. If a DocumentImple
     * object couldn't find for a docid, then the String of this docid was
     * added to vetor rather than DocumentImple object.
     *
     * @param docIdList, a vetor hold a docid list for a data package. In
     *            docid, t here is version number in it.
     */
    private Vector getOldVersionAllDocumentImpl(Vector docIdList)
    {
        //Connection dbConn=null;
        Vector documentImplList = new Vector();
        String siteCode = null;
        String uniqueId = null;
        int rev = 0;

        // Check the parameter
        if (docIdList.isEmpty()) { return documentImplList; }//if

        //for every docid in vector
        for (int i = 0; i < docIdList.size(); i++) {

            String docidPlusVersion = (String) (docIdList.elementAt(i));

            try {
                //create new documentImpl object
                DocumentImpl documentImplObject = new DocumentImpl(
                        docidPlusVersion);
                //add them to vector
                documentImplList.add(documentImplObject);
            }//try
            catch (McdbDocNotFoundException notFoundE) {
                MetaCatUtil.debugMessage(
                        "Error in DBQuery.getOldVersionAllDocument" + "Imple"
                                + notFoundE.getMessage(), 30);
                // Rather than add a DocumentImple object into vetor, a String
                // object
                // - the doicd was added to the vector
                documentImplList.add(docidPlusVersion);
                // Continue the for loop
                continue;
            }//catch
            catch (Exception e) {
                MetaCatUtil.debugMessage(
                        "Error in DBQuery.getOldVersionAllDocument" + "Imple"
                                + e.getMessage(), 30);
                // Continue the for loop
                continue;
            }//catch

        }//for
        return documentImplList;
    }//getOldVersionAllDocumentImple

    /**
     * put a data file into a zip output stream
     *
     * @param docImpl, docmentImpl object which will be sent to zip output
     *            stream
     * @param zipOut, the zip output stream which the docImpl will be put
     * @param packageZipEntry, the zip entry name for whole package
     */
    private void addDataFileToZipOutputStream(DocumentImpl docImpl,
            ZipOutputStream zipOut, String packageZipEntry)
            throws ClassNotFoundException, IOException, SQLException,
            McdbException, Exception
    {
        byte[] byteString = null;
        ZipEntry zEntry = null;
        // this is data file; add file to zip
        String filePath = MetaCatUtil.getOption("datafilepath");
        if (!filePath.endsWith("/")) {
            filePath += "/";
        }
        String fileName = filePath + docImpl.getDocID();
        zEntry = new ZipEntry(packageZipEntry + "/data/" + docImpl.getDocID());
        zipOut.putNextEntry(zEntry);
        FileInputStream fin = null;
        try {
            fin = new FileInputStream(fileName);
            byte[] buf = new byte[4 * 1024]; // 4K buffer
            int b = fin.read(buf);
            while (b != -1) {
                zipOut.write(buf, 0, b);
                b = fin.read(buf);
            }//while
            zipOut.closeEntry();
        }//try
        catch (IOException ioe) {
            MetaCatUtil.debugMessage("There is an exception: "
                    + ioe.getMessage(), 30);
        }//catch
    }//addDataFileToZipOutputStream()

    /**
     * create a html summary for data package and put it into zip output stream
     *
     * @param docImplList, the documentImpl ojbects in data package
     * @param zipOut, the zip output stream which the html should be put
     * @param packageZipEntry, the zip entry name for whole package
     */
    private void addHtmlSummaryToZipOutputStream(Vector docImplList,
            ZipOutputStream zipOut, String packageZipEntry) throws Exception
    {
        StringBuffer htmlDoc = new StringBuffer();
        ZipEntry zEntry = null;
        byte[] byteString = null;
        InputStream source;
        DBTransform xmlToHtml;

        //create a DBTransform ojbect
        xmlToHtml = new DBTransform();
        //head of html
        htmlDoc.append("<html><head></head><body>");
        for (int i = 0; i < docImplList.size(); i++) {
            // If this String object, this means it is missed data file
            if ((((docImplList.elementAt(i)).getClass()).toString())
                    .equals("class java.lang.String")) {

                htmlDoc.append("<a href=\"");
                String dataFileid = (String) docImplList.elementAt(i);
                htmlDoc.append("./data/").append(dataFileid).append("\">");
                htmlDoc.append("Data File: ");
                htmlDoc.append(dataFileid).append("</a><br>");
                htmlDoc.append("<br><hr><br>");

            }//if
            else if ((((DocumentImpl) docImplList.elementAt(i)).getDoctype())
                    .compareTo("BIN") != 0) { //this is an xml file so we can
                                              // transform it.
                //transform each file individually then concatenate all of the
                //transformations together.

                //for metadata xml title
                htmlDoc.append("<h2>");
                htmlDoc.append(((DocumentImpl) docImplList.elementAt(i))
                        .getDocID());
                //htmlDoc.append(".");
                //htmlDoc.append(((DocumentImpl)docImplList.elementAt(i)).getRev());
                htmlDoc.append("</h2>");
                //do the actual transform
                StringWriter docString = new StringWriter();
                xmlToHtml.transformXMLDocument(((DocumentImpl) docImplList
                        .elementAt(i)).toString(), "-//NCEAS//eml-generic//EN",
                        "-//W3C//HTML//EN", "html", docString);
                htmlDoc.append(docString.toString());
                htmlDoc.append("<br><br><hr><br><br>");
            }//if
            else { //this is a data file so we should link to it in the html
                htmlDoc.append("<a href=\"");
                String dataFileid = ((DocumentImpl) docImplList.elementAt(i))
                        .getDocID();
                htmlDoc.append("./data/").append(dataFileid).append("\">");
                htmlDoc.append("Data File: ");
                htmlDoc.append(dataFileid).append("</a><br>");
                htmlDoc.append("<br><hr><br>");
            }//else
        }//for
        htmlDoc.append("</body></html>");
        byteString = htmlDoc.toString().getBytes();
        zEntry = new ZipEntry(packageZipEntry + "/metadata.html");
        zEntry.setSize(byteString.length);
        zipOut.putNextEntry(zEntry);
        zipOut.write(byteString, 0, byteString.length);
        zipOut.closeEntry();
        //dbConn.close();

    }//addHtmlSummaryToZipOutputStream

    /**
     * put a data packadge into a zip output stream
     *
     * @param docId, which the user want to put into zip output stream
     * @param out, a servletoutput stream which the zip output stream will be
     *            put
     * @param user, the username of the user
     * @param groups, the group of the user
     */
    public ZipOutputStream getZippedPackage(String docIdString,
            ServletOutputStream out, String user, String[] groups,
            String passWord) throws ClassNotFoundException, IOException,
            SQLException, McdbException, NumberFormatException, Exception
    {
        ZipOutputStream zOut = null;
        String elementDocid = null;
        DocumentImpl docImpls = null;
        //Connection dbConn = null;
        Vector docIdList = new Vector();
        Vector documentImplList = new Vector();
        Vector htmlDocumentImplList = new Vector();
        String packageId = null;
        String rootName = "package";//the package zip entry name

        String docId = null;
        int version = -5;
        // Docid without revision
        docId = MetaCatUtil.getDocIdFromString(docIdString);
        // revision number
        version = MetaCatUtil.getVersionFromString(docIdString);

        //check if the reqused docId is a data package id
        if (!isDataPackageId(docId)) {

            /*
             * Exception e = new Exception("The request the doc id "
             * +docIdString+ " is not a data package id");
             */

            //CB 1/6/03: if the requested docid is not a datapackage, we just
            // zip
            //up the single document and return the zip file.
            if (!hasPermissionToExportPackage(docId, user, groups)) {

                Exception e = new Exception("User " + user
                        + " does not have permission"
                        + " to export the data package " + docIdString);
                throw e;
            }

            docImpls = new DocumentImpl(docId);
            //checking if the user has the permission to read the documents
            if (DocumentImpl.hasReadPermission(user, groups, docImpls
                    .getDocID())) {
                zOut = new ZipOutputStream(out);
                //if the docImpls is metadata
                if ((docImpls.getDoctype()).compareTo("BIN") != 0) {
                    //add metadata into zip output stream
                    addDocToZipOutputStream(docImpls, zOut, rootName);
                }//if
                else {
                    //it is data file
                    addDataFileToZipOutputStream(docImpls, zOut, rootName);
                    htmlDocumentImplList.add(docImpls);
                }//else
            }//if

            zOut.finish(); //terminate the zip file
            return zOut;
        }
        // Check the permission of user
        else if (!hasPermissionToExportPackage(docId, user, groups)) {

            Exception e = new Exception("User " + user
                    + " does not have permission"
                    + " to export the data package " + docIdString);
            throw e;
        } else //it is a packadge id
        {
            //store the package id
            packageId = docId;
            //get current version in database
            int currentVersion = getCurrentRevFromXMLDoumentsTable(packageId);
            //If it is for current version (-1 means user didn't specify
            // revision)
            if ((version == -1) || version == currentVersion) {
                //get current version number
                version = currentVersion;
                //get package zip entry name
                //it should be docId.revsion.package
                rootName = packageId + MetaCatUtil.getOption("accNumSeparator")
                        + version + MetaCatUtil.getOption("accNumSeparator")
                        + "package";
                //get the whole id list for data packadge
                docIdList = getCurrentDocidListForDataPackage(packageId);
                //get the whole documentImple object
                documentImplList = getCurrentAllDocumentImpl(docIdList);

            }//if
            else if (version > currentVersion || version < -1) {
                throw new Exception("The user specified docid: " + docId + "."
                        + version + " doesn't exist");
            }//else if
            else //for an old version
            {

                rootName = docIdString
                        + MetaCatUtil.getOption("accNumSeparator") + "package";
                //get the whole id list for data packadge
                docIdList = getOldVersionDocidListForDataPackage(docIdString);

                //get the whole documentImple object
                documentImplList = getOldVersionAllDocumentImpl(docIdList);
            }//else

            // Make sure documentImplist is not empty
            if (documentImplList.isEmpty()) { throw new Exception(
                    "Couldn't find component for data package: " + packageId); }//if

            zOut = new ZipOutputStream(out);
            //put every element into zip output stream
            for (int i = 0; i < documentImplList.size(); i++) {
                // if the object in the vetor is String, this means we couldn't
                // find
                // the document locally, we need find it remote
                if ((((documentImplList.elementAt(i)).getClass()).toString())
                        .equals("class java.lang.String")) {
                    // Get String object from vetor
                    String documentId = (String) documentImplList.elementAt(i);
                    MetaCatUtil.debugMessage("docid: " + documentId, 30);
                    // Get doicd without revision
                    String docidWithoutRevision = MetaCatUtil
                            .getDocIdFromString(documentId);
                    MetaCatUtil.debugMessage("docidWithoutRevsion: "
                            + docidWithoutRevision, 30);
                    // Get revision
                    String revision = MetaCatUtil
                            .getRevisionStringFromString(documentId);
                    MetaCatUtil.debugMessage("revsion from docIdentifier: "
                            + revision, 30);
                    // Zip entry string
                    String zipEntryPath = rootName + "/data/";
                    // Create a RemoteDocument object
                    RemoteDocument remoteDoc = new RemoteDocument(
                            docidWithoutRevision, revision, user, passWord,
                            zipEntryPath);
                    // Here we only read data file from remote metacat
                    String docType = remoteDoc.getDocType();
                    if (docType != null) {
                        if (docType.equals("BIN")) {
                            // Put remote document to zip output
                            remoteDoc.readDocumentFromRemoteServerByZip(zOut);
                            // Add String object to htmlDocumentImplList
                            String elementInHtmlList = remoteDoc
                                    .getDocIdWithoutRevsion()
                                    + MetaCatUtil.getOption("accNumSeparator")
                                    + remoteDoc.getRevision();
                            htmlDocumentImplList.add(elementInHtmlList);
                        }//if
                    }//if

                }//if
                else {
                    //create a docmentImpls object (represent xml doc) base on
                    // the docId
                    docImpls = (DocumentImpl) documentImplList.elementAt(i);
                    //checking if the user has the permission to read the
                    // documents
                    if (DocumentImpl.hasReadPermission(user, groups, docImpls
                            .getDocID())) {
                        //if the docImpls is metadata
                        if ((docImpls.getDoctype()).compareTo("BIN") != 0) {
                            //add metadata into zip output stream
                            addDocToZipOutputStream(docImpls, zOut, rootName);
                            //add the documentImpl into the vetor which will
                            // be used in html
                            htmlDocumentImplList.add(docImpls);

                        }//if
                        else {
                            //it is data file
                            addDataFileToZipOutputStream(docImpls, zOut,
                                    rootName);
                            htmlDocumentImplList.add(docImpls);
                        }//else
                    }//if
                }//else
            }//for

            //add html summary file
            addHtmlSummaryToZipOutputStream(htmlDocumentImplList, zOut,
                    rootName);
            zOut.finish(); //terminate the zip file
            //dbConn.close();
            return zOut;
        }//else
    }//getZippedPackage()

    private class ReturnFieldValue
    {

        private String docid = null; //return field value for this docid

        private String fieldValue = null;

        private String xmlFieldValue = null; //return field value in xml
                                             // format

        public void setDocid(String myDocid)
        {
            docid = myDocid;
        }

        public String getDocid()
        {
            return docid;
        }

        public void setFieldValue(String myValue)
        {
            fieldValue = myValue;
        }

        public String getFieldValue()
        {
            return fieldValue;
        }

        public void setXMLFieldValue(String xml)
        {
            xmlFieldValue = xml;
        }

        public String getXMLFieldValue()
        {
            return xmlFieldValue;
        }

    }
}
