/**
 *  '$RCSfile: DBTransform.java,v $'
 *    Purpose: A Class that transforms an XML text document
 *             into a another type using XSL
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: 1.5.0
 *
 * '$Author: tao $'
 * '$Date: 2004/04/01 01:48:54 $'
 * '$Revision: 1.28 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.sql.*;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Stack;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Attr;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.DocumentType;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;
import org.apache.xerces.dom.DocumentTypeImpl;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.NamedNodeMap;

/*
import oracle.xml.parser.v2.XSLStylesheet;
import oracle.xml.parser.v2.XSLException;
import oracle.xml.parser.v2.XMLParseException;
import oracle.xml.parser.v2.XSLProcessor;
import oracle.xml.parser.v2.XMLDocument;
import oracle.xml.parser.v2.DOMParser;
*/
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

/**
 * A Class that transforms XML documents utitlizing XSL style sheets
 */
public class DBTransform {

  //private Connection	conn = null;
  private MetaCatUtil   util = null;
  private String 	configDir = null;
  private String	defaultStyle = null;

  /**
   * construct a DBTransform instance.
   *
   * Generally, one calls transformXMLDocument() after constructing the instance
   *
   * @param conn the database connection from which to lookup the public ids
   */
  public DBTransform()
                  throws IOException,
                         SQLException,
                         ClassNotFoundException
  {
    //this.conn = conn;
    util = new MetaCatUtil();
    configDir = util.getOption("config-dir");
    defaultStyle = util.getOption("default-style");
  }

  /**
   * @see transformXMLDocument(String doc, String sourceType,
   *            String targetType, String qformat, PrintWriter pw,
   *            String sessionid)
   */
  public void transformXMLDocument(String doc, String sourceType,
                String targetType, String qformat, PrintWriter pw,
                Hashtable param)
  {
    transformXMLDocument(doc, sourceType, targetType, qformat, pw, param, null);
  }

   /**
   * @see transformXMLDocument(String doc, String sourceType,
   *            String targetType, String qFormat, StringWriter pw
   *            String sessionid)
   */
  public void transformXMLDocument(String doc, String sourceType,
                String targetType, String qFormat, StringWriter pw)
  {
    transformXMLDocument(doc, sourceType, targetType, qFormat, pw, null);
  }


  /**
   * Transform an XML document using the stylesheet reference from the db
   *
   * @param doc the document to be transformed
   * @param sourcetype the document type of the source
   * @param targettype the target document type
   * @param qformat the name of the style set to use
   * @param pw the PrintWriter to which output is printed
   * @param params some parameters for eml2 transformation
   */
  public void transformXMLDocument(String doc, String sourceType,
                                   String targetType, String qformat,
                                   PrintWriter pw, Hashtable param,
                                   String sessionid)
 {

    // Look up the stylesheet for this type combination
    String xslSystemId = getStyleSystemId(qformat, sourceType, targetType);
    if (xslSystemId != null)
    {
      try
      {// Create a stylesheet from the system id that was found
        doc = removeDOCTYPE(doc);
        StringReader xml = new StringReader(doc);
        StreamResult result = new StreamResult(pw);
        doTransform(xml, result, xslSystemId, param, qformat, sessionid);

      }
      catch (Exception e)
      {
        pw.println(xslSystemId + "Error transforming document in " +
                   "DBTransform.transformXMLDocument: " +
                   e.getMessage());

      }
    }
    else
    {
      // No stylesheet registered form this document type, so just return the
      // XML stream we were passed
      pw.print(doc);
    }
  }

  /**
   * Transform an XML document to StringWriter using the stylesheet reference
   * from the db
   * @param doc the document to be transformed
   * @param sourceType the document type of the source
   * @param targetType the target document type
   * @param qFormat the name of the style set to use
   * @param pw the StringWriter to which output will be stored
   */
  public void transformXMLDocument(String doc, String sourceType,
                String targetType, String qFormat, StringWriter pw,
                String sessionid)
  {

    // Look up the stylesheet for this type combination
    String xslSystemId = getStyleSystemId(qFormat, sourceType, targetType);
    if (xslSystemId != null)
    {
      // Create a stylesheet from the system id that was found
      try
      {
        doc = removeDOCTYPE(doc);
        StringReader xml = new StringReader(doc);
        StreamResult result = new StreamResult(pw);
        doTransform(xml, result, xslSystemId, null, qFormat, sessionid);
      }
      catch (Exception e)
      {
        util.debugMessage(xslSystemId + "Error transforming document in " +
                   "DBTransform.transformXMLDocument: " +
                   e.getMessage(), 30);

      }
    }
    else
    {
      // No stylesheet registered form this document type, so just return the
      // XML stream we were passed
      pw.write(doc);
    }
  }

  /**
   * Method to do transform for a string reader
   * @param doc the document to be transformed
   * @param sourcetype the document type of the source
   * @param targettype the target document type
   * @param qformat the name of the style set to use
   * @param pw the PrintWriter to which output is printed
   * @param params some parameters for eml2 transformation
   */
   public void transformXMLDocument(StringReader docContent, String sourceType,
                                   String targetType, String qformat,
                                   PrintWriter pw, Hashtable param,
                                   String sessionid)
   {
     // Look up the stylesheet for this type combination
    String xslSystemId = getStyleSystemId(qformat, sourceType, targetType);
    if (xslSystemId != null)
    {
      try
      {// Create a stylesheet from the system id that was found
        StreamResult result = new StreamResult(pw);
        doTransform(docContent, result, xslSystemId, param, qformat, sessionid);

      }
      catch (Exception e)
      {
        pw.println(xslSystemId + "Error transforming document in " +
                   "DBTransform.transformXMLDocument: " +
                   e.getMessage());

      }
    }
    else
    {
      // No stylesheet registered form this document type, so just return the
      // XML stream we were passed
      pw.print(docContent);
    }
   }

  /*
   * Method to do transfer
   */
  private void doTransform(StringReader docContent, StreamResult resultOutput,
                           String xslSystemId, Hashtable param,
                           String qformat, String sessionid)
                          throws Exception
  {
    if (xslSystemId != null)
    {
        TransformerFactory tFactory = TransformerFactory.newInstance();
        Transformer transformer = tFactory.newTransformer(
                                  new StreamSource(xslSystemId));
        transformer.setParameter("qformat", qformat);
        MetaCatUtil.debugMessage("qformat: "+qformat, 30);

        if(sessionid != null)
        {
          transformer.setParameter("sessid", sessionid);
        }
        // Set up parameter for transformation
        if ( param != null)
        {
          Enumeration en = param.keys();
          while (en.hasMoreElements())
          {
            String key =(String)en.nextElement();
            String value = ((String[])(param.get(key)))[0];
            MetaCatUtil.debugMessage(key+" : "+value, 30);
            transformer.setParameter(key, value);
          }
        }
        StreamSource xml = new StreamSource(docContent);
        transformer.transform(xml,resultOutput);
    }
  }//doTransform


  /**
   * gets the content of a tag in a given xml file with the given path
   * @param f the file to parse
   * @param path the path to get the content from
   */
  public static NodeList getPathContent(File f, String path)
  {
    if(f == null)
    {
      return null;
    }

    DOMParser parser = new DOMParser();
    InputSource in;
    FileInputStream fs;

    try
    {
      fs = new FileInputStream(f);
      in = new InputSource(fs);
    }
    catch(FileNotFoundException fnf)
    {
      fnf.printStackTrace();
      return null;
    }

    try
    {
      parser.parse(in);
      fs.close();
    }
    catch(Exception e1)
    {
      System.err.println("File: " + f.getPath() + " : parse threw: " +
                         e1.toString());
      return null;
    }

    Document doc = parser.getDocument();

    try
    {
      NodeList docNodeList = XPathAPI.selectNodeList(doc, path);
      return docNodeList;
    }
    catch(Exception se)
    {
      System.err.println("file: " + f.getPath() + " : parse threw: " +
                         se.toString());
      return null;
    }
  }

  /**
   * Lookup a stylesheet reference from the db catalog
   *
   * @param qformat    the named style-set format
   * @param sourcetype the document type of the source
   * @param targettype the document type of the target
   */
  public String getStyleSystemId(String qformat, String sourcetype,
                String targettype) {
    String systemId = null;

    if ((qformat == null) || (qformat.equals("html"))) {
      qformat = defaultStyle;
    }

    // Load the style-set map for this qformat into a DOM
    try {
      boolean breakflag = false;
      String filename = configDir + "/" + qformat + "/" + qformat + ".xml";
      util.debugMessage("Trying style-set file: " + filename, 30);
      File f = new File(filename);
      NodeList nlDoctype = getPathContent(f, "/style-set/doctype");
      NodeList nlDefault = getPathContent(f, "/style-set/default-style");
      Node nDefault = nlDefault.item(0);
      systemId = nDefault.getFirstChild().getNodeValue(); //set the default

      for(int i=0; i<nlDoctype.getLength(); i++)
      { //look for the right sourcetype
        Node nDoctype = nlDoctype.item(i);
        NamedNodeMap atts = nDoctype.getAttributes();
        Node nAtt = atts.getNamedItem("publicid");
        String doctype = nAtt.getFirstChild().getNodeValue();
        if(doctype.equals(sourcetype))
        { //found the right sourcetype now we need to get the target type
          NodeList nlChildren = nDoctype.getChildNodes();
          for(int j=0; j<nlChildren.getLength(); j++)
          {
            Node nChild = nlChildren.item(j);
            String childName = nChild.getNodeName();
            if(childName.equals("target"))
            {
              NamedNodeMap childAtts = nChild.getAttributes();
              Node nTargetPublicId = childAtts.getNamedItem("publicid");
              String target = nTargetPublicId.getFirstChild().getNodeValue();
              if(target.equals(targettype))
              { //we found the right target type
                NodeList nlTarget = nChild.getChildNodes();
                for(int k=0; k<nlTarget.getLength(); k++)
                {
                  Node nChildText = nlTarget.item(k);
                  if(nChildText.getNodeType() == Node.TEXT_NODE)
                  { //get the text from the target node
                    systemId = nChildText.getNodeValue();
                    breakflag = true;
                    break;
                  }
                }
              }
            }

            if(breakflag)
            {
              break;
            }
          }
        }

        if(breakflag)
        {
          break;
        }
      }
    }
    catch(Exception e)
    {
      System.out.println("Error parsing style-set file: " + e.getMessage());
      e.printStackTrace();
    }

    // Return the system ID for this particular source document type
    MetaCatUtil.debugMessage("style system id is: "+systemId, 30);
    return systemId;
  }

 /* Method to modified the system id of xml input -- make sure it
    points to system id in xml_catalog table
  */
  private void modifiedXmlStreamSource(StreamSource xml, String publicId)
                                       throws Exception
  {
    // make sure the xml is not null
    if (xml == null || publicId == null)
    {
      return;
    }
    MetaCatUtil.debugMessage("public id of input stream is " +publicId, 25);
    // Get system id from xml_catalog table
    String systemId = DBEntityResolver.getDTDSystemID(publicId);
    MetaCatUtil.debugMessage("system id of input stream from xml_catalog"
                               +"table is " +systemId, 25);
    //set system id to input stream
    xml.setSystemId(systemId);
  }

  /*
   * removes the DOCTYPE element and its contents from a Sting
   * used to avoid problems with incorrect SystemIDs
   */
  private String removeDOCTYPE(String in) {
    String ret = "";
    int startpos = in.indexOf("<!DOCTYPE");
    if (startpos>-1) {
      int stoppos = in.indexOf(">", startpos + 8);
      ret = in.substring(0,startpos) + in.substring(stoppos+1,in.length());
    } else {
      return in;
    }
    return ret;
  }

  /**
   * the main routine used to test the transform utility.
   *
   * Usage: java DBTransform
   */
  static public void main(String[] args) {

     if (args.length > 0)
     {
        System.err.println("Wrong number of arguments!!!");
        System.err.println("USAGE: java DBTransform");
        return;
     } else {
        try {

          // Open a connection to the database
          /*MetaCatUtil   util = new MetaCatUtil();
          Connection dbconn = util.openDBConnection();*/

          // Create a test document
          StringBuffer testdoc = new StringBuffer();
          testdoc.append("<?xml version=\"1.0\"?>");
          testdoc.append("<eml-dataset><metafile_id>NCEAS-0001</metafile_id>");
          testdoc.append("<dataset_id>DS001</dataset_id>");
          testdoc.append("<title>My test doc</title></eml-dataset>");

          // Transform the document to the new doctype
          DBTransform dbt = new DBTransform();
          dbt.transformXMLDocument(testdoc.toString(),
                                   "-//NCEAS//eml-dataset//EN",
                                   "-//W3C//HTML//EN",
                                   "knb",
                                   new PrintWriter(System.out), null);

        } catch (Exception e) {
          System.err.println("EXCEPTION HANDLING REQUIRED");
          System.err.println(e.getMessage());
          e.printStackTrace(System.err);
        }
     }
  }

  private void dbg(int position) {
    System.err.println("Debug flag: " + position);
  }

}
