/**
 *  '$RCSfile: MetacatReplication.java,v $'
 *    Purpose: A Class that implements replication for metacat
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Chad Berkley
 *    Release: 1.5.0
 *
 *   '$Author: tao $'
 *     '$Date: 2004/09/17 00:04:08 $'
 * '$Revision: 1.67 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import edu.ucsb.nceas.dbadapter.AbstractDatabase;
import java.util.*;
import java.io.*;
import java.sql.*;
import java.net.*;
import java.lang.*;
import java.text.*;
import javax.servlet.*;
import javax.servlet.http.*;

import org.xml.sax.*;

public class MetacatReplication extends HttpServlet implements Runnable
{
  private String deltaT;
  Timer replicationDaemon;
  private static MetaCatUtil util = new MetaCatUtil();
  private Vector fileLocks = new Vector();
  private Thread lockThread = null;
  private static final AbstractDatabase dbAdapter = MetaCatUtil.dbAdapter;
  public static final String FORCEREPLICATEDELETE = "forcereplicatedelete";

  /**
   * Initialize the servlet by creating appropriate database connections
   */
  public void init(ServletConfig config) throws ServletException
  {
     //initialize db connections to handle any update requests
    MetaCatUtil util = new MetaCatUtil();
    deltaT = util.getOption("deltaT");
    //the default deltaT can be set from metacat.properties
    //create a thread to do the delta-T check but don't execute it yet
    replicationDaemon = new Timer(true);
    /*java.net.URL.setURLStreamHandlerFactory(
       new java.net.URLStreamHandlerFactory()
       {
          public java.net.URLStreamHandler createURLStreamHandler(
                        final String protocol)
          {
              if ("http".equals(protocol))
              {
                 try
                 {
                    URLStreamHandler urlsh = new HTTPClient.http.Handler();
                    MetaCatUtil.debugMessage("Using HttpClient Protocol Handler for http", 45);
                    return urlsh;
                  }
                  catch (Exception e)
                  {
                     MetaCatUtil.debugMessage(
                          "Error setting URL StreamHandler!", 30);
                     return null;
                  }
               }//if
               else if ("https".equals(protocol))
               {
                 try
                 {
                   URLStreamHandler urlsh = new HTTPClient.https.Handler();
                   MetaCatUtil.debugMessage("Using HttpClient Protocol Handler for https", 45);
                   return urlsh;
                 }
                 catch (Exception e)
                 {
                   MetaCatUtil.debugMessage(
                              "Error setting URL StreamHandler!", 30);
                   return null;
                 }
              }
              return null;
         }//createURLStreamHandler
     });*/

  }

  public void destroy()
  {
    replicationDaemon.cancel();
   // System.out.println("Replication daemon cancelled.");
  }

  public void doGet (HttpServletRequest request, HttpServletResponse response)
                     throws ServletException, IOException
  {
    // Process the data and send back the response
    handleGetOrPost(request, response);
  }

  public void doPost(HttpServletRequest request, HttpServletResponse response)
                     throws ServletException, IOException
  {
    // Process the data and send back the response
    handleGetOrPost(request, response);
  }

  private void handleGetOrPost(HttpServletRequest request,
                               HttpServletResponse response)
                               throws ServletException, IOException
  {
    //PrintWriter out = response.getWriter();
    //ServletOutputStream outPut = response.getOutputStream();
    Hashtable params = new Hashtable();
    Enumeration paramlist = request.getParameterNames();



// NOT NEEDED - doesn't provide enough security because of possible IP spoofing
// REPLACED with running replication comminications over HTTPS
//    String requestingServerIP = request.getRemoteAddr();
//    InetAddress iaddr = InetAddress.getByName(requestingServerIP);
//    String requestingServer = iaddr.getHostName();

    while (paramlist.hasMoreElements()) {
      String name = (String)paramlist.nextElement();
      String[] value = request.getParameterValues(name);
      params.put(name, value);
    }

    String action = ((String[])params.get("action"))[0];
    String server = null;

    try {
      // check if the server is included in the list of replicated servers
      if ( !action.equals("servercontrol") &&
           !action.equals("stop") &&
           !action.equals("start") &&
           !action.equals("getall") ) {

        server = ((String[])params.get("server"))[0];
        if ( getServerCodeForServerName(server) == 0 ) {
          System.out.println("Action \"" + action +
                             "\" rejected for server: " + server);
          return;
        } else {
          System.out.println("Action \"" + action +
                             "\" accepted for server: " + server);
        }
      }
    } catch (Exception e) {
      System.out.println("Error in MetacatReplication.handleGetOrPost: " +
                         e.getMessage() );
      return;
    }
    if ( action.equals("readdata") )
    {
      OutputStream out=response.getOutputStream();
      //to get the data file.
      handleGetDataFileRequest(out, params, response);
      out.close();
    }
    else if ( action.equals("forcereplicatedatafile") )
    {
      //read a specific docid from remote host, and store it into local host
      handleForceReplicateDataFileRequest(params);

    }
    else
    {
    PrintWriter out = response.getWriter();
    if ( action.equals("stop") ) {
      //stop the replication server
      replicationDaemon.cancel();
      replicationDaemon = new Timer(true);
      out.println("Replication Handler Stopped");
      MetacatReplication.replLog("deltaT handler stopped");


    } else if ( action.equals("start") ) {

      //start the replication server
      int rate;
      if ( params.containsKey("rate") ) {
        rate = new Integer(
               new String(((String[])params.get("rate"))[0])).intValue();
        if(rate < 30) {
            out.println("Replication deltaT rate cannot be less than 30!");
            //deltaT<30 is a timing mess!
            rate = 1000;
        }
      } else {
        rate = 1000;
      }
      out.println("New rate is: " + rate + " seconds.");
      replicationDaemon.cancel();
      replicationDaemon = new Timer(true);
      replicationDaemon.scheduleAtFixedRate(new ReplicationHandler(out), 0,
                                            rate * 1000);
      out.println("Replication Handler Started");
      MetacatReplication.replLog("deltaT handler started with rate=" +
                                    rate + " seconds");


    } else if ( action.equals("getall") ) {
      //updates this server exactly once
      replicationDaemon.schedule(new ReplicationHandler(out), 0);
      response.setContentType("text/html");
      out.println("<html><body>\"Get All\" Done</body></html>");

    } else if ( action.equals("forcereplicate") ) {
      //read a specific docid from remote host, and store it into local host
      handleForceReplicateRequest(out, params, response);

    } else if ( action.equals(FORCEREPLICATEDELETE) ) {
      //read a specific docid from remote host, and store it into local host
      handleForceReplicateDeleteRequest(out, params, response);

    } else if ( action.equals("update") ) {
      //request an update list from the server
      handleUpdateRequest(out, params, response);

    } else if ( action.equals("read") ) {
      //request a specific document from the server
      //note that this could be replaced by a call to metacatServlet
      //handleGetDocumentAction().
      handleGetDocumentRequest(out, params, response);
    } else if ( action.equals("getlock") ) {
      handleGetLockRequest(out, params, response);

    } else if ( action.equals("getdocumentinfo") ) {
      handleGetDocumentInfoRequest(out, params, response);

    } else if ( action.equals("gettime") ) {
      handleGetTimeRequest(out, params, response);

    } else if ( action.equals("getcatalog") ) {
      handleGetCatalogRequest(out, params, response, true);

    } else if ( action.equals("servercontrol") ) {
      handleServerControlRequest(out, params, response);
    } else if ( action.equals("test") ) {
      response.setContentType("text/html");
      out.println("<html><body>Test successfully</body></html>");
    }

    out.close();
    }//else
  }

  /**
   * This method can add, delete and list the servers currently included in
   * xml_replication.
   * action           subaction            other needed params
   * ---------------------------------------------------------
   * servercontrol    add                  server
   * servercontrol    delete               server
   * servercontrol    list
   */
  private void handleServerControlRequest(PrintWriter out, Hashtable params,
                                          HttpServletResponse response)
  {
    String subaction = ((String[])params.get("subaction"))[0];
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    String replicate =null;
    String server = null;
    String dataReplicate = null;
    String hub = null;
    try {
      //conn = util.openDBConnection();
      dbConn=DBConnectionPool.
               getDBConnection("MetacatReplication.handleServerControlRequest");
      serialNumber=dbConn.getCheckOutSerialNumber();

      // add server to server list
      if ( subaction.equals("add") ) {
        replicate = ((String[])params.get("replicate"))[0];
        server = ((String[])params.get("server"))[0];

        //Get data replication value
        dataReplicate = ((String[])params.get("datareplicate"))[0];
        //Get hub value
        hub = ((String[])params.get("hub"))[0];

        /*pstmt = dbConn.prepareStatement("INSERT INTO xml_replication " +
                  "(server, last_checked, replicate, datareplicate, hub) " +
                                      "VALUES ('" + server + "', to_date(" +
                                      "'01/01/00', 'MM/DD/YY'), '" +
                                      replicate +"', '" +dataReplicate+"', '"
                                      + hub +"')");*/
        pstmt = dbConn.prepareStatement("INSERT INTO xml_replication " +
                  "(server, last_checked, replicate, datareplicate, hub) " +
                                      "VALUES ('" + server + "', "+
                                      dbAdapter.toDate("01/01/1980", "MM/DD/YYYY")
                                      + ", '" +
                                      replicate +"', '" +dataReplicate+"', '"
                                      + hub +"')");

        pstmt.execute();
        pstmt.close();
        dbConn.commit();
        out.println("Server " + server + " added");
        response.setContentType("text/html");
        out.println("<html><body><table border=\"1\">");
        out.println("<tr><td><b>server</b></td><td><b>last_checked</b></td><td>");
        out.println("<b>replicate</b></td>");
        out.println("<td><b>datareplicate</b></td>");
        out.println("<td><b>hub</b></td></tr>");
        pstmt = dbConn.prepareStatement("SELECT * FROM xml_replication");
        //increase dbconnection usage
        dbConn.increaseUsageCount(1);

        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean tablehasrows = rs.next();
        while(tablehasrows) {
          out.println("<tr><td>" + rs.getString(2) + "</td><td>");
          out.println(rs.getString(3) + "</td><td>");
          out.println(rs.getString(4) + "</td><td>");
          out.println(rs.getString(5) + "</td><td>");
          out.println(rs.getString(6) + "</td></tr>");

          tablehasrows = rs.next();
        }
        out.println("</table></body></html>");

        // download certificate with the public key on this server
        // and import it as a trusted certificate
        String certURL = ((String[])params.get("certificate"))[0];
        downloadCertificate(certURL);

      // delete server from server list
      } else if ( subaction.equals("delete") ) {
        server = ((String[])params.get("server"))[0];
        pstmt = dbConn.prepareStatement("DELETE FROM xml_replication " +
                                      "WHERE server LIKE '" + server + "'");
        pstmt.execute();
        pstmt.close();
        dbConn.commit();
        out.println("Server " + server + " deleted");
        response.setContentType("text/html");
        out.println("<html><body><table border=\"1\">");
        out.println("<tr><td><b>server</b></td><td><b>last_checked</b></td><td>");
        out.println("<b>replicate</b></td>");
        out.println("<td><b>datareplicate</b></td>");
        out.println("<td><b>hub</b></td></tr>");

        pstmt = dbConn.prepareStatement("SELECT * FROM xml_replication");
        //increase dbconnection usage
        dbConn.increaseUsageCount(1);
        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean tablehasrows = rs.next();
        while(tablehasrows)
        {
          out.println("<tr><td>" + rs.getString(2) + "</td><td>");
          out.println(rs.getString(3) + "</td><td>");
          out.println(rs.getString(4) + "</td><td>");
          out.println(rs.getString(5) + "</td><td>");
          out.println(rs.getString(6) + "</td></tr>");
          tablehasrows = rs.next();
        }
        out.println("</table></body></html>");

      // list servers in server list
      } else if ( subaction.equals("list") ) {
        response.setContentType("text/html");
        out.println("<html><body><table border=\"1\">");
        out.println("<tr><td><b>server</b></td><td><b>last_checked</b></td><td>");
        out.println("<b>replicate</b></td>");
        out.println("<td><b>datareplicate</b></td>");
        out.println("<td><b>hub</b></td></tr>");
        pstmt = dbConn.prepareStatement("SELECT * FROM xml_replication");
        pstmt.execute();
        ResultSet rs = pstmt.getResultSet();
        boolean tablehasrows = rs.next();
        while(tablehasrows) {
          out.println("<tr><td>" + rs.getString(2) + "</td><td>");
          out.println(rs.getString(3) + "</td><td>");
          out.println(rs.getString(4) + "</td><td>");
          out.println(rs.getString(5) + "</td><td>");
          out.println(rs.getString(6) + "</td></tr>");
          tablehasrows = rs.next();
        }
        out.println("</table></body></html>");
      }
      else
      {

        out.println("<error>Unkonwn subaction</error>");

      }
      pstmt.close();
      //conn.close();

    } catch(Exception e) {
      System.out.println("Error in " +
                         "MetacatReplication.handleServerControlRequest " +
                         e.getMessage());
      e.printStackTrace(System.out);
    }
    finally
    {
      try
      {
        pstmt.close();
      }//try
      catch (SQLException ee)
      {
        MetaCatUtil.debugMessage("Error in " +
                "MetacatReplication.handleServerControlRequest to close pstmt "
                 + ee.getMessage(), 30);
      }//catch
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//finally
    }//finally

  }

  // download certificate with the public key from certURL and
  // upload it onto this server; it then must be imported as a
  // trusted certificate
  private void downloadCertificate (String certURL)
                throws FileNotFoundException, IOException, MalformedURLException
  {
    MetaCatUtil util = new MetaCatUtil();
    String certPath = util.getOption("certPath"); //the path to be uploaded to

    // get filename from the URL of the certificate
    String filename = certURL;
    int slash = Math.max(filename.lastIndexOf('/'), filename.lastIndexOf('\\'));
    if ( slash > -1 ) {
      filename = filename.substring(slash + 1);
    }

    // open file output strem to write the input into it
    File f = new File(certPath, filename);
    synchronized (f) {
      try {
        if ( f.exists() ) {
          throw new IOException("File already exist: " + f.getCanonicalFile());
          //if ( f.exists() && !f.canWrite() ) {
          //  throw new IOException("Not writable: " + f.getCanonicalFile());
        }
      } catch (SecurityException se) {
        // if a security manager exists,
        // its checkRead method is called for f.exist()
        // or checkWrite method is called for f.canWrite()
        throw se;
      }

      // create a buffered byte output stream
      // that uses a default-sized output buffer
      FileOutputStream fos = new FileOutputStream(f);
      BufferedOutputStream out = new BufferedOutputStream(fos);

      // this should be http url
      URL url = new URL(certURL);
      BufferedInputStream bis = null;
      try {
        bis = new BufferedInputStream(url.openStream());
        byte[] buf = new byte[4 * 1024]; // 4K buffer
        int b = bis.read(buf);
        while (b != -1) {
          out.write(buf, 0, b);
          b = bis.read(buf);
        }
      } finally {
        if (bis != null) bis.close();
      }
      // the input and the output streams must be closed
      bis.close();
            out.flush();
            out.close();
            fos.close();
    } // end of synchronized(f)
  }

  /**
   * when a forcereplication request comes in, local host sends a read request
   * to the requesting server (remote server) for the specified docid.
   * Then store it in local database.
   */
  private void handleForceReplicateRequest(PrintWriter out, Hashtable params,
                                           HttpServletResponse response)
  {
    String server = ((String[])params.get("server"))[0]; // the server that
    String docid = ((String[])params.get("docid"))[0]; // sent the document
    String dbaction = "UPDATE"; // the default action is UPDATE
    boolean override = false;
    int serverCode = 1;
    DBConnection dbConn = null;
    int serialNumber = -1;

    try {
      //if the url contains a dbaction then the default action is overridden
      if(params.containsKey("dbaction")) {
        dbaction = ((String[])params.get("dbaction"))[0];
        //serverCode = MetacatReplication.getServerCode(server);
        //override = true; //we are now overriding the default action
      }
      MetacatReplication.replLog("force replication request from " + server);
      MetaCatUtil.debugMessage("Force replication request from: "+ server, 34);
      MetaCatUtil.debugMessage("Force replication docid: "+docid, 34);
      MetaCatUtil.debugMessage("Force replication action: "+dbaction, 34);
      // sending back read request to remote server
      URL u = new URL("https://" + server + "?server="
                +util.getLocalReplicationServerName()
                +"&action=read&docid=" + docid);
      String xmldoc = MetacatReplication.getURLContent(u);

      // get the document info from server
      URL docinfourl = new URL("https://" + server +
                               "?server="+util.getLocalReplicationServerName()
                               +"&action=getdocumentinfo&docid=" + docid);

      String docInfoStr = MetacatReplication.getURLContent(docinfourl);

      //dih is the parser for the docinfo xml format
      DocInfoHandler dih = new DocInfoHandler();
      XMLReader docinfoParser = ReplicationHandler.initParser(dih);
      docinfoParser.parse(new InputSource(new StringReader(docInfoStr)));
      Hashtable docinfoHash = dih.getDocInfo();

      // Get user owner of this docid
      String user = (String)docinfoHash.get("user_owner");
      // Get home server of this docid
      String homeServer=(String)docinfoHash.get("home_server");
      MetaCatUtil.debugMessage("homeServer: "+homeServer, 34);
      // Get Document type
      String docType = (String)docinfoHash.get("doctype");
      MetaCatUtil.debugMessage("docType: "+docType, 34);
      String parserBase = null;
      // this for eml2 and we need user eml2 parser
      if (docType != null &&
          (docType.trim()).equals(DocumentImpl.EML2_0_0NAMESPACE))
      {
         MetaCatUtil.debugMessage("This is an eml200 document!", 30);
         parserBase = DocumentImpl.EML200;
      }
      else if (docType != null && (docType.trim()).equals(DocumentImpl.EML2_0_1NAMESPACE))
      {
         MetaCatUtil.debugMessage("This is an eml2.0.1 document!", 30);
         parserBase = DocumentImpl.EML200;
      }
      else if (docType != null && (docType.trim()).equals(DocumentImpl.EML2_1_0NAMESPACE))
      {
         MetaCatUtil.debugMessage("This is an eml2.1.0 document!", 30);
         parserBase = DocumentImpl.EML210;
      }
      MetaCatUtil.debugMessage("The parserBase is: "+parserBase, 30);

      // Get DBConnection from pool
      dbConn=DBConnectionPool.
              getDBConnection("MetacatReplication.handleForceReplicateRequest");
      serialNumber=dbConn.getCheckOutSerialNumber();
      // write the document to local database
      DocumentImplWrapper wrapper = new DocumentImplWrapper(parserBase, false);
      wrapper.writeReplication(dbConn, new StringReader(xmldoc), null, null,
                               dbaction, docid, user, null, homeServer, server);

      MetacatReplication.replLog("document " + docid + " added to DB with " +
                                 "action " + dbaction);
    }//try
    catch(Exception e)
    {
      MetacatReplication.replErrorLog("document " + docid +
                                      " failed to added to DB with " +
                                      "action " + dbaction + " because "+
                                       e.getMessage());
      MetaCatUtil.debugMessage("ERROR in MetacatReplication.handleForceReplicate" +
                         "Request(): " + e.getMessage(), 30);

    }//catch
    finally
    {
      // Return the checked out DBConnection
      DBConnectionPool.returnDBConnection(dbConn, serialNumber);
    }//finally
  }

/*
 * when a forcereplication delete request comes in, local host will delete this
 * document
 */
private void handleForceReplicateDeleteRequest(PrintWriter out, Hashtable params,
                                         HttpServletResponse response)
{
  String server = ((String[])params.get("server"))[0]; // the server that
  String docid = ((String[])params.get("docid"))[0]; // sent the document
  try
  {
    MetacatReplication.replLog("force replication delete request from " + server);
    MetacatReplication.replLog("force replication delete docid " + docid);
    MetaCatUtil.debugMessage("Force replication delete request from: "+ server, 34);
    MetaCatUtil.debugMessage("Force replication delete docid: "+docid, 34);
    DocumentImpl.delete(docid, null, null, server);
    MetacatReplication.replLog("document " + docid + " was successfully deleted ");
    MetaCatUtil.debugMessage("document " + docid + " was successfully deleted ", 34);
  }
  catch(Exception e)
  {
    MetacatReplication.replErrorLog("document " + docid +
                                    " failed to delete because "+
                                     e.getMessage());
    MetaCatUtil.debugMessage("ERROR in MetacatReplication.handleForceDeleteReplicate" +
                       "Request(): " + e.getMessage(), 30);

  }//catch

}


  /**
   * when a forcereplication data file request comes in, local host sends a
   * readdata request to the requesting server (remote server) for the specified
   * docid. Then store it in local database and file system
   */
  private void handleForceReplicateDataFileRequest(Hashtable params)
  {

    //make sure there is some parameters
    if(params.isEmpty())
    {
      return;
    }
    // Get remote server
    String server = ((String[])params.get("server"))[0];
    // the docid should include rev number
    String docid = ((String[])params.get("docid"))[0];
    // Make sure there is a docid and server
    if (docid==null || server==null || server.equals(""))
    {
      MetaCatUtil.debugMessage("Didn't specify docid or server for replication"
                              , 20);
      return;
    }

    // Overide or not
    boolean override = false;
    // dbaction - update or insert
    String dbaction=null;

    try
    {
      //docid was switch to two parts uinque code and rev
      String uniqueCode=MetaCatUtil.getDocIdFromString(docid);
      int rev=MetaCatUtil.getVersionFromString(docid);
      if(params.containsKey("dbaction"))
      {
        dbaction = ((String[])params.get("dbaction"))[0];
      }
      else//default value is update
      {
        dbaction = "update";
      }

      MetacatReplication.replLog("force replication request from " + server);
      MetaCatUtil.debugMessage("Force replication request from: "+ server, 34);
      MetaCatUtil.debugMessage("Force replication docid: "+docid, 34);
      MetaCatUtil.debugMessage("Force replication action: "+dbaction, 34);
      // get the document info from server
      URL docinfourl = new URL("https://" + server +
                               "?server="+util.getLocalReplicationServerName()
                               +"&action=getdocumentinfo&docid=" + uniqueCode);

      String docInfoStr = MetacatReplication.getURLContent(docinfourl);

      //dih is the parser for the docinfo xml format
      DocInfoHandler dih = new DocInfoHandler();
      XMLReader docinfoParser = ReplicationHandler.initParser(dih);
      docinfoParser.parse(new InputSource(new StringReader(docInfoStr)));
      Hashtable docinfoHash = dih.getDocInfo();
      String user = (String)docinfoHash.get("user_owner");

      String docName = (String)docinfoHash.get("docname");

      String docType = (String)docinfoHash.get("doctype");

      String docHomeServer= (String)docinfoHash.get("home_server");
      MetaCatUtil.debugMessage("docHomeServer of datafile: "+docHomeServer, 34);



      //if action is delete, we don't delete the data file. Just archieve
      //the xml_documents
      /*if (dbaction.equals("delete"))
      {
        //conn = util.getConnection();
        DocumentImpl.delete(docid,user,null);
        //util.returnConnection(conn);
      }*/
      //To data file insert or update is same
      if (dbaction.equals("insert")||dbaction.equals("update"))
      {
        //Get data file and store it into local file system.
        // sending back readdata request to server
        URL url = new URL("https://" + server + "?server="
                +util.getLocalReplicationServerName()
                +"&action=readdata&docid=" + docid);
        String datafilePath = util.getOption("datafilepath");
        //register data file into xml_documents table and wite data file
        //into file system
        DocumentImpl.writeDataFileInReplication(url.openStream(), datafilePath,
                            docName, docType, docid, user,docHomeServer,server);
     }



    MetacatReplication.replLog("datafile " + docid + " added to DB with " +
                                 "action " + dbaction);
    }
    catch(Exception e)
    {

      MetacatReplication.replErrorLog("Datafile " + docid +
                                      " failed to added to DB with " +
                                      "action " + dbaction + " because "+
                                       e.getMessage());
      MetaCatUtil.debugMessage
              ("ERROR in MetacatReplication.handleForceDataFileReplicate" +
                         "Request(): " + e.getMessage(), 30);
    }
  }
  /**
   * Grants or denies a lock to a requesting host.
   * The servlet parameters of interrest are:
   * docid: the docid of the file the lock is being requested for
   * currentdate: the timestamp of the document on the remote server
   *
   */
  private void handleGetLockRequest(PrintWriter out, Hashtable params,
                                    HttpServletResponse response)
  {

    try
    {

      String docid = ((String[])params.get("docid"))[0];
      String remoteRev = ((String[])params.get("updaterev"))[0];
      DocumentImpl requestDoc = new DocumentImpl(docid);
      MetacatReplication.replLog("lock request for " + docid);
      int localRevInt = requestDoc.getRev();
      int remoteRevInt = Integer.parseInt(remoteRev);

      if(remoteRevInt >= localRevInt)
      {
        if(!fileLocks.contains(docid))
        { //grant the lock if it is not already locked
          fileLocks.add(0, docid); //insert at the beginning of the queue Vector
          //send a message back to the the remote host authorizing the insert
          out.println("<lockgranted><docid>" +docid+ "</docid></lockgranted>");
          lockThread = new Thread(this);
          lockThread.setPriority(Thread.MIN_PRIORITY);
          lockThread.start();
          MetacatReplication.replLog("lock granted for " + docid);
        }
        else
        { //deny the lock
          out.println("<filelocked><docid>" + docid + "</docid></filelocked>");
          MetacatReplication.replLog("lock denied for " + docid +
                                     "reason: file already locked");
        }
      }
      else
      {//deny the lock.
        out.println("<outdatedfile><docid>" + docid + "</docid></filelocked>");
        MetacatReplication.replLog("lock denied for " + docid +
                                   "reason: client has outdated file");
      }
      //conn.close();
    }
    catch(Exception e)
    {
      System.out.println("error requesting file lock from MetacatReplication." +
                         "handleGetLockRequest: " + e.getMessage());
      e.printStackTrace(System.out);
    }
  }

  /**
   * Sends all of the xml_documents information encoded in xml to a requestor
   * the format is:
   * <!ELEMENT documentinfo (docid, docname, doctype, doctitle, user_owner,
   *                  user_updated, home_server, public_access, rev)/>
   * all of the subelements of document info are #PCDATA
   */
  private void handleGetDocumentInfoRequest(PrintWriter out, Hashtable params,
                                        HttpServletResponse response)
  {
    String docid = ((String[])(params.get("docid")))[0];
    StringBuffer sb = new StringBuffer();

    try
    {

      DocumentImpl doc = new DocumentImpl(docid);
      sb.append("<documentinfo><docid>").append(docid);
      sb.append("</docid><docname>").append(doc.getDocname());
      sb.append("</docname><doctype>").append(doc.getDoctype());
      sb.append("</doctype>");
      sb.append("<user_owner>").append(doc.getUserowner());
      sb.append("</user_owner><user_updated>").append(doc.getUserupdated());
      sb.append("</user_updated>");
      sb.append("<home_server>");
      sb.append(doc.getDocHomeServer());
      sb.append("</home_server>");
      sb.append("<public_access>").append(doc.getPublicaccess());
      sb.append("</public_access><rev>").append(doc.getRev());
      sb.append("</rev></documentinfo>");
      response.setContentType("text/xml");
      out.println(sb.toString());

    }
    catch (Exception e)
    {
      System.out.println("error in " +
                         "metacatReplication.handlegetdocumentinforequest: " +
                          e.getMessage());
    }

  }

  /**
   * Sends a datafile to a remote host
   */
  private void handleGetDataFileRequest(OutputStream outPut,
                            Hashtable params, HttpServletResponse response)

  {
    // File path for data file
    String filepath = util.getOption("datafilepath");
    // Request docid
    String docId = ((String[])(params.get("docid")))[0];
    //check if the doicd is null
    if (docId==null)
    {
      util.debugMessage("Didn't specify docid for replication", 20);
      return;
    }

    //try to open a https stream to test if the request server's public key
    //in the key store, this is security issue
    try
    {
      String server = ((String[])params.get("server"))[0];
      URL u = new URL("https://" + server + "?server="
                +util.getLocalReplicationServerName()
                +"&action=test");
      String test = MetacatReplication.getURLContent(u);
      //couldn't pass the test
      if (test.indexOf("successfully")==-1)
      {
        //response.setContentType("text/xml");
        //outPut.println("<error>Couldn't pass the trust test</error>");
        MetaCatUtil.debugMessage("Couldn't pass the trust test", 20);
        return;
      }
    }//try
    catch (Exception ee)
    {
      return;
    }//catch

    if(!filepath.endsWith("/"))
    {
          filepath += "/";
    }
    // Get file aboslute file name
    String filename = filepath + docId;

    //MIME type
    String contentType = null;
    if (filename.endsWith(".xml"))
    {
        contentType="text/xml";
    }
    else if (filename.endsWith(".css"))
    {
        contentType="text/css";
    }
    else if (filename.endsWith(".dtd"))
    {
        contentType="text/plain";
    }
    else if (filename.endsWith(".xsd"))
    {
        contentType="text/xml";
    }
    else if (filename.endsWith("/"))
    {
        contentType="text/html";
    }
    else
    {
        File f = new File(filename);
        if ( f.isDirectory() )
        {
           contentType="text/html";
        }
        else
        {
           contentType="application/octet-stream";
        }
     }

   // Set the mime type
   response.setContentType(contentType);

   // Get the content of the file
   FileInputStream fin = null;
   try
   {
      // FileInputStream to metacat
      fin = new FileInputStream(filename);
      // 4K buffer
      byte[] buf = new byte[4 * 1024];
      // Read data from file input stream to byte array
      int b = fin.read(buf);
      // Write to outStream from byte array
      while (b != -1)
      {
        outPut.write(buf, 0, b);
        b = fin.read(buf);
      }
      // close file input stream
      fin.close();

   }//try
   catch(Exception e)
   {
      System.out.println("error getting data file from MetacatReplication." +
                         "handlGetDataFileRequest " + e.getMessage());
      e.printStackTrace(System.out);
   }//catch

}


  /**
   * Sends a document to a remote host
   */
  private void handleGetDocumentRequest(PrintWriter out, Hashtable params,
                                        HttpServletResponse response)
  {

    try
    {
      //try to open a https stream to test if the request server's public key
      //in the key store, this is security issue
      String server = ((String[])params.get("server"))[0];
      URL u = new URL("https://" + server + "?server="
                +util.getLocalReplicationServerName()
                +"&action=test");
      String test = MetacatReplication.getURLContent(u);
      //couldn't pass the test
      if (test.indexOf("successfully")==-1)
      {
        response.setContentType("text/xml");
        out.println("<error>Couldn't pass the trust test</error>");
        out.close();
        return;
      }

      String docid = ((String[])(params.get("docid")))[0];

      DocumentImpl di = new DocumentImpl(docid);
      response.setContentType("text/xml");
      out.print(di.toString(null, null, true));

      MetacatReplication.replLog("document " + docid + " sent");

    }
    catch(Exception e)
    {
      MetaCatUtil.debugMessage("error getting document from MetacatReplication."
                          +"handlGetDocumentRequest " + e.getMessage(), 30);
      //e.printStackTrace(System.out);
      response.setContentType("text/xml");
      out.println("<error>"+e.getMessage()+"</error>");
    }

  }

  /**
   * Sends a list of all of the documents on this sever along with their
   * revision numbers.
   * The format is:
   * <!ELEMENT replication (server, updates)>
   * <!ELEMENT server (#PCDATA)>
   * <!ELEMENT updates ((updatedDocument | deleteDocument)*)>
   * <!ELEMENT updatedDocument (docid, rev, datafile*)>
   * <!ELEMENT deletedDocument (docid, rev)>
   * <!ELEMENT docid (#PCDATA)>
   * <!ELEMENT rev (#PCDATA)>
   * <!ELEMENT datafile (#PCDATA)>
   * note that the rev in deletedDocument is always empty.  I just left
   * it in there to make the parser implementation easier.
   */
  private void handleUpdateRequest(PrintWriter out, Hashtable params,
                                    HttpServletResponse response)
  {
    // Checked out DBConnection
    DBConnection dbConn = null;
    // DBConenction serial number when checked it out
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    // Server list to store server info of xml_replication table
    ReplicationServerList serverList = null;

    try
    {
      // Check out a DBConnection from pool
      dbConn=DBConnectionPool.
                  getDBConnection("MetacatReplication.handleUpdateRequest");
      serialNumber=dbConn.getCheckOutSerialNumber();
      // Create a server list from xml_replication table
      serverList = new ReplicationServerList();

      // Get remote server name from param
      String server = ((String[])params.get("server"))[0];
      // If no servr name in param, return a error
      if ( server == null || server.equals(""))
      {
        response.setContentType("text/xml");
        out.println("<error>Request didn't specify server name</error>");
        out.close();
        return;
      }//if

      //try to open a https stream to test if the request server's public key
      //in the key store, this is security issue
      URL u = new URL("https://" + server + "?server="
                +util.getLocalReplicationServerName()
                +"&action=test");
      String test = MetacatReplication.getURLContent(u);
      //couldn't pass the test
      if (test.indexOf("successfully")==-1)
      {
        response.setContentType("text/xml");
        out.println("<error>Couldn't pass the trust test</error>");
        out.close();
        return;
      }


      // Check if local host configure to replicate xml documents to remote
      // server. If not send back a error message
      if (!serverList.getReplicationValue(server))
      {
        response.setContentType("text/xml");
        out.println
        ("<error>Configuration not allow to replicate document to you</error>");
        out.close();
        return;
      }//if

      // Store the sql command
      StringBuffer docsql = new StringBuffer();
      // Stroe the docid list
      StringBuffer doclist = new StringBuffer();
      // Store the deleted docid list
      StringBuffer delsql = new StringBuffer();
      // Store the data set file
      Vector packageFiles = new Vector();

      // Append local server's name and replication servlet to doclist
      doclist.append("<?xml version=\"1.0\"?><replication>");
      doclist.append("<server>").append(util.getOption("server"));
      doclist.append(util.getOption("replicationpath"));
      doclist.append("</server><updates>");

      // Get correct docid that reside on this server according the requesting
      // server's replicate and data replicate value in xml_replication table
      docsql.append("select docid, rev, doctype from xml_documents ");
      // If the localhost is not a hub to the remote server, only replicate
      // the docid' which home server is local host (server_location =1)
      if (!serverList.getHubValue(server))
      {
        docsql.append("where server_location = 1");
      }
      MetaCatUtil.debugMessage("Doc sql: "+docsql.toString(), 30);

      // Get any deleted documents
      delsql.append("select distinct docid from ");
      delsql.append("xml_revisions where docid not in (select docid from ");
      delsql.append("xml_documents) ");
      // If the localhost is not a hub to the remote server, only replicate
      // the docid' which home server is local host (server_location =1)
      if (!serverList.getHubValue(server))
      {
        delsql.append("and server_location = 1");
      }
      MetaCatUtil.debugMessage("Deleted sql: "+delsql.toString(), 30);



      // Get docid list of local host
      pstmt = dbConn.prepareStatement(docsql.toString());
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      //If metacat configed to replicate data file
      //if ((util.getOption("replicationsenddata")).equals("on"))
      if (serverList.getDataReplicationValue(server))
      {
        while(tablehasrows)
        {
          String recordDoctype = rs.getString(3);
          Vector packagedoctypes = MetaCatUtil.getOptionList(
                                     MetaCatUtil.getOption("packagedoctype"));
          //if this is a package file, put it at the end
          //because if a package file is read before all of the files it
          //refers to are loaded then there is an error
          if(recordDoctype != null && !packagedoctypes.contains(recordDoctype))
          {
              //If this is not data file
              if (!recordDoctype.equals("BIN"))
              {
                //for non-data file document
                doclist.append("<updatedDocument>");
                doclist.append("<docid>").append(rs.getString(1));
                doclist.append("</docid><rev>").append(rs.getInt(2));
                doclist.append("</rev>");
                doclist.append("</updatedDocument>");
              }//if
              else
              {
                //for data file document, in datafile attributes
                //we put "datafile" value there
                doclist.append("<updatedDocument>");
                doclist.append("<docid>").append(rs.getString(1));
                doclist.append("</docid><rev>").append(rs.getInt(2));
                doclist.append("</rev>");
                doclist.append("<datafile>");
                doclist.append(MetaCatUtil.getOption("datafileflag"));
                doclist.append("</datafile>");
                doclist.append("</updatedDocument>");
              }//else
          }//if packagedoctpes
          else
          { //the package files are saved to be put into the xml later.
              Vector v = new Vector();
              v.add(new String(rs.getString(1)));
              v.add(new Integer(rs.getInt(2)));
              packageFiles.add(new Vector(v));
          }//esle
          tablehasrows = rs.next();
        }//while
      }//if
      else //metacat was configured not to send data file
      {
        while(tablehasrows)
        {
          String recordDoctype = rs.getString(3);
          if(!recordDoctype.equals("BIN"))
          { //don't replicate data files
            Vector packagedoctypes = MetaCatUtil.getOptionList(
                                     MetaCatUtil.getOption("packagedoctype"));
            if(recordDoctype != null && !packagedoctypes.contains(recordDoctype))
            {   //if this is a package file, put it at the end
              //because if a package file is read before all of the files it
              //refers to are loaded then there is an error
              doclist.append("<updatedDocument>");
              doclist.append("<docid>").append(rs.getString(1));
              doclist.append("</docid><rev>").append(rs.getInt(2));
              doclist.append("</rev>");
              doclist.append("</updatedDocument>");
            }
            else
            { //the package files are saved to be put into the xml later.
              Vector v = new Vector();
              v.add(new String(rs.getString(1)));
              v.add(new Integer(rs.getInt(2)));
              packageFiles.add(new Vector(v));
            }
         }//if
         tablehasrows = rs.next();
        }//while
      }//else

      pstmt = dbConn.prepareStatement(delsql.toString());
      //usage count should increas 1
      dbConn.increaseUsageCount(1);

      pstmt.execute();
      rs = pstmt.getResultSet();
      tablehasrows = rs.next();
      while(tablehasrows)
      { //handle the deleted documents
        doclist.append("<deletedDocument><docid>").append(rs.getString(1));
        doclist.append("</docid><rev></rev></deletedDocument>");
        //note that rev is always empty for deleted docs
        tablehasrows = rs.next();
      }

      //now we can put the package files into the xml results
      for(int i=0; i<packageFiles.size(); i++)
      {
        Vector v = (Vector)packageFiles.elementAt(i);
        doclist.append("<updatedDocument>");
        doclist.append("<docid>").append((String)v.elementAt(0));
        doclist.append("</docid><rev>");
        doclist.append(((Integer)v.elementAt(1)).intValue());
        doclist.append("</rev>");
        doclist.append("</updatedDocument>");
      }

      doclist.append("</updates></replication>");
      MetaCatUtil.debugMessage("doclist: " + doclist.toString(), 40);
      pstmt.close();
      //conn.close();
      response.setContentType("text/xml");
      out.println(doclist.toString());

    }
    catch(Exception e)
    {
      MetaCatUtil.debugMessage("error in MetacatReplication." +
                         "handleupdaterequest: " + e.getMessage(), 30);
      //e.printStackTrace(System.out);
      response.setContentType("text/xml");
      out.println("<error>"+e.getMessage()+"</error>");
    }
    finally
    {
      try
      {
        pstmt.close();
      }//try
      catch (SQLException ee)
      {
        MetaCatUtil.debugMessage("Error in MetacatReplication." +
                "handleUpdaterequest to close pstmt: "+ee.getMessage(), 30);
      }//catch
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//finally
    }//finally

  }//handlUpdateRequest

  /**
   * Returns the xml_catalog table encoded in xml
   */
  public static String getCatalogXML()
  {
    return handleGetCatalogRequest(null, null, null, false);
  }

  /**
   * Sends the contents of the xml_catalog table encoded in xml
   * The xml format is:
   * <!ELEMENT xml_catalog (row*)>
   * <!ELEMENT row (entry_type, source_doctype, target_doctype, public_id,
   *                system_id)>
   * All of the sub elements of row are #PCDATA

   * If printFlag == false then do not print to out.
   */
  private static String handleGetCatalogRequest(PrintWriter out,
                                                Hashtable params,
                                                HttpServletResponse response,
                                                boolean printFlag)
  {
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    try
    {
      /*conn = MetacatReplication.getDBConnection("MetacatReplication." +
                                                "handleGetCatalogRequest");*/
      dbConn=DBConnectionPool.
                 getDBConnection("MetacatReplication.handleGetCatalogRequest");
      serialNumber=dbConn.getCheckOutSerialNumber();
      pstmt = dbConn.prepareStatement("select entry_type, " +
                              "source_doctype, target_doctype, public_id, " +
                              "system_id from xml_catalog");
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      StringBuffer sb = new StringBuffer();
      sb.append("<?xml version=\"1.0\"?><xml_catalog>");
      while(tablehasrows)
      {
        sb.append("<row><entry_type>").append(rs.getString(1));
        sb.append("</entry_type><source_doctype>").append(rs.getString(2));
        sb.append("</source_doctype><target_doctype>").append(rs.getString(3));
        sb.append("</target_doctype><public_id>").append(rs.getString(4));
        sb.append("</public_id><system_id>").append(rs.getString(5));
        sb.append("</system_id></row>");

        tablehasrows = rs.next();
      }
      sb.append("</xml_catalog>");
      //conn.close();
      if(printFlag)
      {
        response.setContentType("text/xml");
        out.println(sb.toString());
      }
      pstmt.close();
      return sb.toString();
    }
    catch(Exception e)
    {

      System.out.println("error in MetacatReplication.handleGetCatalogRequest:"+
                          e.getMessage());
      e.printStackTrace(System.out);
      if(printFlag)
      {
        out.println("<error>"+e.getMessage()+"</error>");
      }
    }
    finally
    {
      try
      {
        pstmt.close();
      }//try
      catch (SQLException ee)
      {
        MetaCatUtil.
           debugMessage("Error in MetacatReplication.handleGetCatalogRequest: "
           +ee.getMessage(), 30);
      }//catch
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//finally
    }//finally

    return null;
  }

  /**
   * Sends the current system date to the remote server.  Using this action
   * for replication gets rid of any problems with syncronizing clocks
   * because a time specific to a document is always kept on its home server.
   */
  private void handleGetTimeRequest(PrintWriter out, Hashtable params,
                                    HttpServletResponse response)
  {
    SimpleDateFormat formatter = new SimpleDateFormat ("MM/dd/yy HH:mm:ss");
    java.util.Date localtime = new java.util.Date();
    String dateString = formatter.format(localtime);
    response.setContentType("text/xml");

    out.println("<timestamp>" + dateString + "</timestamp>");
  }

  /**
   * this method handles the timeout for a file lock.  when a lock is
   * granted it is granted for 30 seconds.  When this thread runs out
   * it deletes the docid from the queue, thus eliminating the lock.
   */
  public void run()
  {
    try
    {
      MetaCatUtil.debugMessage("thread started for docid: " +
                               (String)fileLocks.elementAt(0), 45);

      Thread.sleep(30000); //the lock will expire in 30 seconds
      MetaCatUtil.debugMessage("thread for docid: " +
                             (String)fileLocks.elementAt(fileLocks.size() - 1) +
                              " exiting.", 45);

      fileLocks.remove(fileLocks.size() - 1);
      //fileLocks is treated as a FIFO queue.  If there are more than one lock
      //in the vector, the first one inserted will be removed.
    }
    catch(Exception e)
    {
      MetaCatUtil.debugMessage("error in file lock thread from " +
                                "MetacatReplication.run: " + e.getMessage(), 30);
    }
  }

  /**
   * Returns the name of a server given a serverCode
   * @param serverCode the serverid of the server
   * @return the servername or null if the specified serverCode does not
   *         exist.
   */
  public static String getServerNameForServerCode(int serverCode)
  {
    //System.out.println("serverid: " + serverCode);
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    try
    {
      dbConn=DBConnectionPool.
                  getDBConnection("MetacatReplication.getServer");
      serialNumber=dbConn.getCheckOutSerialNumber();
      String sql = new String("select server from " +
                              "xml_replication where serverid = " +
                              serverCode);
      pstmt = dbConn.prepareStatement(sql);
      //System.out.println("getserver sql: " + sql);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      if(tablehasrows)
      {
        //System.out.println("server: " + rs.getString(1));
        return rs.getString(1);
      }

      //conn.close();
    }
    catch(Exception e)
    {
      System.out.println("Error in MetacatReplication.getServer: " +
                          e.getMessage());
    }
    finally
    {
      try
      {
        pstmt.close();
      }//try
      catch (SQLException ee)
      {
        MetaCatUtil.debugMessage("Error in MetacactReplication.getserver: "+
                                    ee.getMessage(), 30);
      }//catch
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//fianlly
    }//finally



    return null;
      //return null if the server does not exist
  }

  /**
   * Returns a server code given a server name
   * @param server the name of the server
   * @return integer > 0 representing the code of the server, 0 if the server
   *  does not exist.
   */
  public static int getServerCodeForServerName(String server) throws Exception
  {
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    int serverCode = 0;

    try {

      //conn = util.openDBConnection();
      dbConn=DBConnectionPool.
                  getDBConnection("MetacatReplication.getServerCode");
      serialNumber=dbConn.getCheckOutSerialNumber();
      pstmt = dbConn.prepareStatement("SELECT serverid FROM xml_replication " +
                                    "WHERE server LIKE '" + server + "'");
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      if ( tablehasrows ) {
        serverCode = rs.getInt(1);
        pstmt.close();
        //conn.close();
        return serverCode;
      }

    } catch(Exception e) {
      throw e;

    } finally {
      try
      {
        pstmt.close();
        //conn.close();
       }//try
       catch(Exception ee)
       {
         MetaCatUtil.debugMessage("Error in MetacatReplicatio.getServerCode: "
                                  +ee.getMessage(), 30);

       }//catch
       finally
       {
         DBConnectionPool.returnDBConnection(dbConn, serialNumber);
       }//finally
    }//finally

    return serverCode;
  }

  /**
   * Method to get a host server information for given docid
   * @param conn a connection to the database
   */
  public static Hashtable getHomeServerInfoForDocId(String docId)
  {
    Hashtable sl = new Hashtable();
    DBConnection dbConn = null;
    int serialNumber = -1;
    //MetaCatUtil ut=new MetaCatUtil();
    docId=MetaCatUtil.getDocIdFromString(docId);
    PreparedStatement pstmt=null;
    int serverLocation;
    try
    {
      //get conection
      dbConn=DBConnectionPool.
                  getDBConnection("ReplicationHandler.getHomeServer");
      serialNumber=dbConn.getCheckOutSerialNumber();
      //get a server location from xml_document table
      pstmt=dbConn.prepareStatement("select server_location from xml_documents "
                                            +"where docid = ?");
      pstmt.setString(1, docId);
      pstmt.execute();
      ResultSet serverName = pstmt.getResultSet();
      //get a server location
      if(serverName.next())
      {
        serverLocation=serverName.getInt(1);
        pstmt.close();
      }
      else
      {
        pstmt.close();
        //ut.returnConnection(conn);
        return null;
      }
      pstmt=dbConn.prepareStatement("select server, last_checked, replicate " +
                        "from xml_replication where serverid = ?");
      //increase usage count
      dbConn.increaseUsageCount(1);
      pstmt.setInt(1, serverLocation);
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tableHasRows = rs.next();
      if (tableHasRows)
      {

          String server = rs.getString(1);
          String last_checked = rs.getString(2);
          if(!server.equals("localhost"))
          {
            sl.put(server, last_checked);
          }

      }
      else
      {
        pstmt.close();
        //ut.returnConnection(conn);
        return null;
      }
      pstmt.close();
    }
    catch(Exception e)
    {
      System.out.println("error in replicationHandler.getHomeServer(): " +
                         e.getMessage());
    }
    finally
    {
      try
      {
        pstmt.close();
        //ut.returnConnection(conn);
      }
      catch (Exception ee)
      {
        MetaCatUtil.debugMessage("Eror irn rplicationHandler.getHomeServer() "+
                          "to close pstmt: "+ee.getMessage(), 30);
      }
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }

    }//finally
    return sl;
  }

  /**
   * Returns a home server location  given a accnum
   * @param accNum , given accNum for a document
   *
   */
  public static int getHomeServerCodeForDocId(String accNum) throws Exception
  {
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    int serverCode = 1;
    //MetaCatUtil ut = new MetaCatUtil();
    String docId=MetaCatUtil.getDocIdFromString(accNum);

    try
    {

      // Get DBConnection
      dbConn=DBConnectionPool.
                  getDBConnection("ReplicationHandler.getServerLocation");
      serialNumber=dbConn.getCheckOutSerialNumber();
      pstmt=dbConn.prepareStatement("SELECT server_location FROM xml_documents "
                              + "WHERE docid LIKE '" + docId + "'");
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      //If a document is find, return the server location for it
      if ( tablehasrows )
      {
        serverCode = rs.getInt(1);
        pstmt.close();
        //conn.close();
        return serverCode;
      }
      //if couldn't find in xml_documents table, we think server code is 1
      //(this is new document)
      else
      {
        pstmt.close();
        //conn.close();
        return serverCode;
      }

    }
    catch(Exception e)
    {

      throw e;

    }
    finally
    {
      try
      {
        pstmt.close();
        //conn.close();

      }
      catch(Exception ee)
      {
        MetaCatUtil.debugMessage("Erorr in Replication.getServerLocation "+
                     "to close pstmt"+ee.getMessage(), 30);
      }
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//finally
    }//finally
   //return serverCode;
  }



  /**
   * This method returns the content of a url
   * @param u the url to return the content from
   * @return a string representing the content of the url
   * @throws java.io.IOException
   */
  public static String getURLContent(URL u) throws java.io.IOException
  {
    char istreamChar;
    int istreamInt;
    MetaCatUtil.debugMessage("Before open the stream"+u.toString(), 50);
    InputStream input = u.openStream();
    MetaCatUtil.debugMessage("Afetr open the stream"+u.toString(), 50);
    InputStreamReader istream = new InputStreamReader(input);
    StringBuffer serverResponse = new StringBuffer();
    while((istreamInt = istream.read()) != -1)
    {
      istreamChar = (char)istreamInt;
      serverResponse.append(istreamChar);
    }
    istream.close();
    input.close();

    return serverResponse.toString();
  }

  /**
   * Method for writing replication messages to a log file specified in
   * metacat.properties
   */
  public static void replLog(String message)
  {
    try
    {
      FileOutputStream fos = new FileOutputStream(
                                 util.getOption("replicationlog"), true);
      PrintWriter pw = new PrintWriter(fos);
      SimpleDateFormat formatter = new SimpleDateFormat ("yy-MM-dd HH:mm:ss");
      java.util.Date localtime = new java.util.Date();
      String dateString = formatter.format(localtime);
      dateString += " :: " + message;
      //time stamp each entry
      pw.println(dateString);
      pw.flush();
    }
    catch(Exception e)
    {
      System.out.println("error writing to replication log from " +
                         "MetacatReplication.replLog: " + e.getMessage());
      //e.printStackTrace(System.out);
    }
  }

  /**
   * Method for writing replication messages to a log file specified in
   * metacat.properties
   */
  public static void replErrorLog(String message)
  {
    try
    {
      FileOutputStream fos = new FileOutputStream(
                                 util.getOption("replicationerrorlog"), true);
      PrintWriter pw = new PrintWriter(fos);
      SimpleDateFormat formatter = new SimpleDateFormat ("yy-MM-dd HH:mm:ss");
      java.util.Date localtime = new java.util.Date();
      String dateString = formatter.format(localtime);
      dateString += " :: " + message;
      //time stamp each entry
      pw.println(dateString);
      pw.flush();
    }
    catch(Exception e)
    {
      System.out.println("error writing to replication log from " +
                         "MetacatReplication.replLog: " + e.getMessage());
      //e.printStackTrace(System.out);
    }
  }

  /**
   * Returns true if the replicate field for server in xml_replication is 1.
   * Returns false otherwise
   */
  public static boolean replToServer(String server)
  {
    DBConnection dbConn = null;
    int serialNumber = -1;
    PreparedStatement pstmt = null;
    try
    {
      dbConn=DBConnectionPool.
                  getDBConnection("MetacatReplication.repltoServer");
      serialNumber=dbConn.getCheckOutSerialNumber();
      pstmt = dbConn.prepareStatement("select replicate from " +
                                    "xml_replication where server like '" +
                                     server + "'");
      pstmt.execute();
      ResultSet rs = pstmt.getResultSet();
      boolean tablehasrows = rs.next();
      if(tablehasrows)
      {
        int i = rs.getInt(1);
        if(i == 1)
        {
          pstmt.close();
          //conn.close();
          return true;
        }
        else
        {
          pstmt.close();
          //conn.close();
          return false;
        }
      }
    }
    catch(Exception e)
    {
      System.out.println("error in MetacatReplication.replToServer: " +
                         e.getMessage());
    }
    finally
    {
      try
      {
        pstmt.close();
        //conn.close();
      }//try
      catch(Exception ee)
      {
        MetaCatUtil.debugMessage("Error in MetacatReplication.replToServer: "
                                  +ee.getMessage(), 30);
      }//catch
      finally
      {
        DBConnectionPool.returnDBConnection(dbConn, serialNumber);
      }//finally
    }//finally
    return false;
    //the default if this server does not exist is to not replicate to it.
  }


}
