/**
 *  '$RCSfile: QueryGroup.java,v $'
 *    Purpose: A Class that represents a structured query, and can be
 *             constructed from an XML serialization conforming to
 *             pathquery.dtd. The printSQL() method can be used to print
 *             a SQL serialization of the query.
 *  Copyright: 2000 Regents of the University of California and the
 *             National Center for Ecological Analysis and Synthesis
 *    Authors: Matt Jones
 *    Release: 1.5.0
 *
 *   '$Author: sgarg $'
 *     '$Date: 2005/04/22 21:18:49 $'
 * '$Revision: 1.5 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package edu.ucsb.nceas.metacat;

import edu.ucsb.nceas.dbadapter.*;

import java.io.*;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;
import java.util.Enumeration;

 /** a utility class that represents a group of terms in a query */
public  class QueryGroup {
    private String operator = null;  // indicates how query terms are combined
    private Vector children = null;  // the list of query terms and groups
    private int countPercentageSearchItem = 0;
    /**
     * construct a new QueryGroup
     *
     * @param operator the boolean conector used to connect query terms
     *                    in this query group
     */
    public QueryGroup(String operator) {
      this.operator = operator;
      children = new Vector();
    }

    /**
     * Add a child QueryGroup to this QueryGroup
     *
     * @param qgroup the query group to be added to the list of terms
     */
    public void addChild(QueryGroup qgroup) {
      children.add((Object)qgroup);
    }

    /**
     * Add a child QueryTerm to this QueryGroup
     *
     * @param qterm the query term to be added to the list of terms
     */
    public void addChild(QueryTerm qterm) {
      children.add((Object)qterm);
    }

    /**
     * Retrieve an Enumeration of query terms for this QueryGroup
     */
    public Enumeration getChildren() {
      return children.elements();
    }

    public int getPercentageSymbolCount()
    {
      return countPercentageSearchItem;
    }

    /**
     * create a SQL serialization of the query that this instance represents
     */
    public String printSQL(boolean useXMLIndex) {
      StringBuffer self = new StringBuffer();
      StringBuffer queryString = new StringBuffer();

      boolean first = true;

      Enumeration en= getChildren();
      while (en.hasMoreElements()) {
        Object qobject = en.nextElement();

        if (qobject instanceof QueryGroup) {
          if (first) {
            first = false;
          } else {
             if(!queryString.toString().equals("")){
                    queryString.append(" " + operator + " ");
             }
          }
          QueryGroup qg = (QueryGroup)qobject;
          queryString.append(qg.printSQL(useXMLIndex));
          // count percerntage number
          int count = qg.getPercentageSymbolCount();
          countPercentageSearchItem = countPercentageSearchItem + count;
        } else if (qobject instanceof QueryTerm) {
           QueryTerm qt = (QueryTerm)qobject;
           String termQueryString = qt.printSQL(useXMLIndex);
           if(!(qt.getSearchMode().equals("contains") && qt.getValue().equals("%"))){
               if (first) {
                   first = false;
               } else {
                   if(!queryString.toString().equals("")){
                       queryString.append(" " + operator + " ");
                   }
               }
               queryString.append(termQueryString);
           }
           // count percerntage number
           int count = qt.getPercentageSymbolCount();
           countPercentageSearchItem = countPercentageSearchItem + count;
        } else {
          System.err.println("qobject wrong type: fatal error");
        }
      }

      if(!queryString.toString().equals("")){
          self.append("(");
          self.append(queryString.toString());
          self.append(")");
      }
      return self.toString();
    }

    /**
     * create a String description of the query that this instance represents.
     * This should become a way to get the XML serialization of the query.
     */
    public String toString() {
      StringBuffer self = new StringBuffer();

      self.append("  (Query group operator=" + operator + "\n");
      Enumeration en= getChildren();
      while (en.hasMoreElements()) {
        Object qobject = en.nextElement();
        self.append(qobject);
      }
      self.append("  )\n");
      return self.toString();
    }
  }
