/**
 *  '$RCSfile: HarvesterServlet.java,v $'
 *  Copyright: 2004 University of New Mexico and the 
 *             Regents of the University of California
 *
 *   '$Author: costa $'
 *     '$Date: 2005/01/25 21:57:02 $'
 * '$Revision: 1.1 $'
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

package edu.ucsb.nceas.metacat.harvesterClient;

import javax.servlet.*;
import javax.servlet.http.*;

/**
 * HarvesterServlet class allows Harvester to be run as a background
 * process. This eliminates the need to run Harvester in a terminal window.
 * To activate this servlet, uncomment the HarvesterServlet entry in the
 * appropriate 'lib/web.xml.tomcat*' file.
 * 
 * @author costa
 */
public class HarvesterServlet extends HttpServlet implements Runnable {

  Thread harvesterThread;                // background Harvester thread

  /**
   * Initializes the servlet by starting a separate thread in which to
   * run the Harvester main program.
   */
  public void init() throws ServletException {
    harvesterThread = new Thread(this);
    harvesterThread.setPriority(Thread.MIN_PRIORITY);  // be a good citizen
    harvesterThread.start();
  }

  
  /**
   * Runs the Harvester main program in a separate thread. First sleeps for
   * 30 seconds to give Metacat a chance to fully initialize.
   */
  public void run() {
      String[] args = new String[1];
      String schemaLocation = "eml://ecoinformatics.org/harvestList " +
       "http://cvs.ecoinformatics.org/cvs/cvsweb.cgi/~checkout~" +
       "/metacat/lib/harvester/harvestList.xsd?rev=1.1&content-type=text" +
       "/plain&only_with_tag=HEAD";
      args[0] = schemaLocation;

      try {
        Thread.sleep(30000);    // Sleep 30 seconds before starting Harvester
      }
      catch (InterruptedException e) {
        System.err.println("InterruptedException: " + e.getMessage());
      }

      Harvester.main(args);
  }

  
  /**
   * Stops the Harvester thread when the servlet shuts down.
   */
  public void destroy() {
    harvesterThread.stop();
  }
}
